/*
 * ddrcal.c
 *
 *  Created on: 05.03.2014
 *      Author: thomas
 */

#include <common.h>
#include <asm/io.h>
#include <asm/arch/clock.h>
#include <asm/arch/imx-regs.h>
#include <asm/arch/iomux.h>
#include <asm/arch/mx6-pins.h>
#include <asm/arch/mx6-ddr.h>
#include <asm/errno.h>
#include <asm/gpio.h>
#include <asm/imx-common/iomux-v3.h>
#include <asm/imx-common/boot_mode.h>


#ifndef reg32_write
#define reg32_write writel
#define reg32_read readl
#endif

void reg32setbit(int Reg, int Bit)
{
  int NewValue = (reg32_read(Reg) | (1 << Bit));
  reg32_write(Reg, NewValue);
}

void reg32clrbit(int Reg, int Bit)
{
  int NewValue = (reg32_read(Reg) & (~(1 << Bit)));
  reg32_write(Reg, NewValue);
}

#define MMDC_P0_IPS_BASE_ADDR 0x021b0000
#define MMDC_P1_IPS_BASE_ADDR 0x021b4000

#define MDCTL_OFFSET      0x0000
#define MDPDC_OFFSET      0x0004
#define MAPSR_OFFSET      0x0404
#define MDMISC_OFFSET     0x0018
#define MDREF_OFFSET      0x0020
#define MDSCR_OFFSET      0x001C
#define MPPDCMPR1_OFFSET  0x088C
#define MPSWDAR_OFFSET    0x0894
#define MPRDDLCTL_OFFSET  0x088C
#define MPMUR_OFFSET      0x0848
#define MPDGCTRL0_OFFSET  0x083C
#define MPDGCTRL1_OFFSET  0x0840
#define MPDGHWST0_OFFSET  0x087C
#define MPDGHWST1_OFFSET  0x0880
#define MPDGHWST2_OFFSET  0x0884
#define MPDGHWST3_OFFSET  0x0888
#define MPRDDLHWCTL_OFFSET 0x0860
#define MPWRDLCTL_OFFSET 0x0850
#define MPWRDLHWCTL_OFFSET 0x0864

#define IOMUXC_SW_PAD_CTL_PAD_DRAM_SDQS0 MX6_IOM_DRAM_SDQS0
#define IOMUXC_SW_PAD_CTL_PAD_DRAM_SDQS1 MX6_IOM_DRAM_SDQS1
#define IOMUXC_SW_PAD_CTL_PAD_DRAM_SDQS2 MX6_IOM_DRAM_SDQS2
#define IOMUXC_SW_PAD_CTL_PAD_DRAM_SDQS3 MX6_IOM_DRAM_SDQS3
#define IOMUXC_SW_PAD_CTL_PAD_DRAM_SDQS4 MX6_IOM_DRAM_SDQS4
#define IOMUXC_SW_PAD_CTL_PAD_DRAM_SDQS5 MX6_IOM_DRAM_SDQS5
#define IOMUXC_SW_PAD_CTL_PAD_DRAM_SDQS6 MX6_IOM_DRAM_SDQS6
#define IOMUXC_SW_PAD_CTL_PAD_DRAM_SDQS7 MX6_IOM_DRAM_SDQS7

static inline int setgpio(unsigned gpio, int value)
{
  if (value)
    writel(readl(GPIO6_BASE_ADDR) | (1 << (gpio & 0x1f)), GPIO6_BASE_ADDR);
  else
    writel(readl(GPIO6_BASE_ADDR) & ~(1 << (gpio & 0x1f)), GPIO6_BASE_ADDR);
  return 0;
}

int ddr3cal(void)
{

  int cs0_enable_initial = 0;
  int cs1_enable_initial = 0;
  // int PDDWord = 0x55aaaa55; // original values, these work, but can get slightly better below
  int PDDWord = 0x00FFFF00; // best so far, place into MPPDCMPR1

  // check to see which chip selects are enabled
  cs0_enable_initial = (reg32_read(MMDC_P0_IPS_BASE_ADDR + MDCTL_OFFSET) & 0x80000000) >> 31;
  cs1_enable_initial = (reg32_read(MMDC_P0_IPS_BASE_ADDR + MDCTL_OFFSET) & 0x40000000) >> 30;

  printf("DDR: Power Down... CS0:%d CS1:%d\n", cs0_enable_initial, cs1_enable_initial);


  // disable DDR logic power down timer:
  printf("DDR: MDPDC: %08X -> %08X\n", reg32_read((MMDC_P0_IPS_BASE_ADDR + MDPDC_OFFSET)), reg32_read((MMDC_P0_IPS_BASE_ADDR + MDPDC_OFFSET)) & 0xffff00ff);
  reg32_write((MMDC_P0_IPS_BASE_ADDR + MDPDC_OFFSET), reg32_read((MMDC_P0_IPS_BASE_ADDR + MDPDC_OFFSET)) & 0xffff00ff);


  return 444;
}

void ddr3print(void)
{
  printf("MMDC registers updated from calibration \n");
  printf("\nRead DQS Gating calibration\n");
  printf("MPDGCTRL0 PHY0 (0x021b083c) = 0x%08X\n", reg32_read(0x021b083c));
  printf("MPDGCTRL1 PHY0 (0x021b0840) = 0x%08X\n", reg32_read(0x021b0840));
  printf("MPDGCTRL0 PHY1 (0x021b483c) = 0x%08X\n", reg32_read(0x021b483c));
  printf("MPDGCTRL1 PHY1 (0x021b4840) = 0x%08X\n", reg32_read(0x021b4840));
  printf("\nRead calibration\n");
  printf("MPRDDLCTL PHY0 (0x021b0848) = 0x%08X\n", reg32_read(0x021b0848));
  printf("MPRDDLCTL PHY1 (0x021b4848) = 0x%08X\n", reg32_read(0x021b4848));
  printf("\nWrite calibration\n");
  printf("MPWRDLCTL PHY0 (0x021b0850) = 0x%08X\n", reg32_read(0x021b0850));
  printf("MPWRDLCTL PHY1 (0x021b4850) = 0x%08X\n", reg32_read(0x021b4850));
  printf("\n");
  // registers below are for debugging purposes
  // these print out the upper and lower boundaries captured during read DQS gating calibration
  printf("Status registers, upper and lower bounds, for read DQS gating. \n");
  printf("MPDGHWST0 PHY0 (0x021b087c) = 0x%08X\n", reg32_read(0x021b087c));
  printf("MPDGHWST1 PHY0 (0x021b0880) = 0x%08X\n", reg32_read(0x021b0880));
  printf("MPDGHWST2 PHY0 (0x021b0884) = 0x%08X\n", reg32_read(0x021b0884));
  printf("MPDGHWST3 PHY0 (0x021b0888) = 0x%08X\n", reg32_read(0x021b0888));
  printf("MPDGHWST0 PHY1 (0x021b487c) = 0x%08X\n", reg32_read(0x021b487c));
  printf("MPDGHWST1 PHY1 (0x021b4880) = 0x%08X\n", reg32_read(0x021b4880));
  printf("MPDGHWST2 PHY1 (0x021b4884) = 0x%08X\n", reg32_read(0x021b4884));
  printf("MPDGHWST3 PHY1 (0x021b4888) = 0x%08X\n", reg32_read(0x021b4888));
}

typedef int (*DDRCalibFunction_t) (void);


int ddr_calibration(void)
{
  printf("ddr3print: %p\n", ddr3print);
  printf("ddr3cal: %p\n", ddr3cal);

  DDRCalibFunction_t Target = (void *) (IRAM_BASE_ADDR + IRAM_SIZE/2);
  memcpy(Target, ddr3cal, 4096);
  printf("ddr3cal: call at %p -> ", Target);
  printf(" %d\n", Target());
  // ddr3cal();
  ddr3print();
  return 0;
}
