/*
 * Copyright (C) 2012 Freescale Semiconductor, Inc.
 *
 * Author: Fabio Estevam <fabio.estevam@freescale.com>
 *
 * See file CREDITS for list of people who contributed to this
 * project.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 */

#include <common.h>
#include <asm/io.h>
#include <asm/arch/clock.h>
#include <asm/arch/imx-regs.h>
#include <asm/arch/iomux.h>
#include <asm/arch/mx6q_pins.h>
#include <asm/errno.h>
#include <asm/gpio.h>
#include <asm/imx-common/iomux-v3.h>
#include <asm/imx-common/boot_mode.h>
#include <mmc.h>
#include <fsl_esdhc.h>
#include <miiphy.h>
#include <netdev.h>
#include <asm/imx-common/mxc_i2c.h>
#include <i2c.h>

#include "sha1.h"

#include "ddrcal.h"

DECLARE_GLOBAL_DATA_PTR;

#define I2C_PAD_CTRL  (PAD_CTL_PUS_100K_UP |      \
  PAD_CTL_SPEED_MED | PAD_CTL_DSE_40ohm | PAD_CTL_HYS | \
  PAD_CTL_ODE | PAD_CTL_SRE_FAST)

#define UART_PAD_CTRL  (PAD_CTL_PUS_100K_UP |			\
	PAD_CTL_SPEED_MED | PAD_CTL_DSE_40ohm |			\
	PAD_CTL_SRE_FAST  | PAD_CTL_HYS)

#define USDHC_PAD_CTRL (PAD_CTL_PUS_47K_UP |			\
	PAD_CTL_SPEED_LOW | PAD_CTL_DSE_80ohm |			\
	PAD_CTL_SRE_FAST  | PAD_CTL_HYS)

#define ENET_PAD_CTRL  (PAD_CTL_PUS_100K_UP |			\
	PAD_CTL_SPEED_MED | PAD_CTL_DSE_40ohm | PAD_CTL_HYS)

int dram_init(void)
{
	gd->ram_size = get_ram_size((void *)PHYS_SDRAM, PHYS_SDRAM_SIZE);

	return 0;
}

/* I2C2 PMIC */
struct i2c_pads_info i2c_pad_info1 = {
  .scl = {
    .i2c_mode = MX6_PAD_KEY_COL3__I2C2_SCL | MUX_PAD_CTRL(I2C_PAD_CTRL),
    .gpio_mode = MX6_PAD_KEY_COL3__GPIO_4_12 | MUX_PAD_CTRL(I2C_PAD_CTRL),
    .gp = IMX_GPIO_NR(4, 12)
  },
  .sda = {
    .i2c_mode = MX6_PAD_KEY_ROW3__I2C2_SDA | MUX_PAD_CTRL(I2C_PAD_CTRL),
    .gpio_mode = MX6_PAD_KEY_ROW3__GPIO_4_13 | MUX_PAD_CTRL(I2C_PAD_CTRL),
    .gp = IMX_GPIO_NR(4, 13)
  }
};


iomux_v3_cfg_t const uart1_pads[] = {
	MX6_PAD_CSI0_DAT10__UART1_TXD | MUX_PAD_CTRL(UART_PAD_CTRL),
	MX6_PAD_CSI0_DAT11__UART1_RXD | MUX_PAD_CTRL(UART_PAD_CTRL),
};

iomux_v3_cfg_t const enet_pads[] = {
	MX6_PAD_ENET_MDIO__ENET_MDIO		| MUX_PAD_CTRL(ENET_PAD_CTRL),
	MX6_PAD_ENET_MDC__ENET_MDC		| MUX_PAD_CTRL(ENET_PAD_CTRL),
	MX6_PAD_RGMII_TXC__ENET_RGMII_TXC	| MUX_PAD_CTRL(ENET_PAD_CTRL),
	MX6_PAD_RGMII_TD0__ENET_RGMII_TD0	| MUX_PAD_CTRL(ENET_PAD_CTRL),
	MX6_PAD_RGMII_TD1__ENET_RGMII_TD1	| MUX_PAD_CTRL(ENET_PAD_CTRL),
	MX6_PAD_RGMII_TD2__ENET_RGMII_TD2	| MUX_PAD_CTRL(ENET_PAD_CTRL),
	MX6_PAD_RGMII_TD3__ENET_RGMII_TD3	| MUX_PAD_CTRL(ENET_PAD_CTRL),
	MX6_PAD_RGMII_TX_CTL__RGMII_TX_CTL	| MUX_PAD_CTRL(ENET_PAD_CTRL),
	MX6_PAD_ENET_REF_CLK__ENET_TX_CLK	| MUX_PAD_CTRL(ENET_PAD_CTRL),
	MX6_PAD_RGMII_RXC__ENET_RGMII_RXC	| MUX_PAD_CTRL(ENET_PAD_CTRL),
	MX6_PAD_RGMII_RD0__ENET_RGMII_RD0	| MUX_PAD_CTRL(ENET_PAD_CTRL),
	MX6_PAD_RGMII_RD1__ENET_RGMII_RD1	| MUX_PAD_CTRL(ENET_PAD_CTRL),
	MX6_PAD_RGMII_RD2__ENET_RGMII_RD2	| MUX_PAD_CTRL(ENET_PAD_CTRL),
	MX6_PAD_RGMII_RD3__ENET_RGMII_RD3	| MUX_PAD_CTRL(ENET_PAD_CTRL),
	MX6_PAD_RGMII_RX_CTL__RGMII_RX_CTL	| MUX_PAD_CTRL(ENET_PAD_CTRL),
	/* AR8031 PHY Reset */
	MX6_PAD_ENET_CRS_DV__GPIO_1_25		| MUX_PAD_CTRL(NO_PAD_CTRL),
};

static void setup_iomux_enet(void)
{
	imx_iomux_v3_setup_multiple_pads(enet_pads, ARRAY_SIZE(enet_pads));

	/* Reset AR8031 PHY */
	gpio_direction_output(IMX_GPIO_NR(1, 25) , 0);
	udelay(500);
	gpio_set_value(IMX_GPIO_NR(1, 25), 1);
}

iomux_v3_cfg_t const usdhc2_pads[] = {
	MX6_PAD_SD2_CLK__USDHC2_CLK	| MUX_PAD_CTRL(USDHC_PAD_CTRL),
	MX6_PAD_SD2_CMD__USDHC2_CMD	| MUX_PAD_CTRL(USDHC_PAD_CTRL),
	MX6_PAD_SD2_DAT0__USDHC2_DAT0	| MUX_PAD_CTRL(USDHC_PAD_CTRL),
	MX6_PAD_SD2_DAT1__USDHC2_DAT1	| MUX_PAD_CTRL(USDHC_PAD_CTRL),
	MX6_PAD_SD2_DAT2__USDHC2_DAT2	| MUX_PAD_CTRL(USDHC_PAD_CTRL),
	MX6_PAD_SD2_DAT3__USDHC2_DAT3	| MUX_PAD_CTRL(USDHC_PAD_CTRL),
	MX6_PAD_NANDF_D4__USDHC2_DAT4	| MUX_PAD_CTRL(USDHC_PAD_CTRL),
	MX6_PAD_NANDF_D5__USDHC2_DAT5	| MUX_PAD_CTRL(USDHC_PAD_CTRL),
	MX6_PAD_NANDF_D6__USDHC2_DAT6	| MUX_PAD_CTRL(USDHC_PAD_CTRL),
	MX6_PAD_NANDF_D7__USDHC2_DAT7	| MUX_PAD_CTRL(USDHC_PAD_CTRL),
	MX6_PAD_NANDF_D2__GPIO_2_2	| MUX_PAD_CTRL(NO_PAD_CTRL), /* CD */
};

iomux_v3_cfg_t const usdhc3_pads[] = {
	MX6_PAD_SD3_CLK__USDHC3_CLK   | MUX_PAD_CTRL(USDHC_PAD_CTRL),
	MX6_PAD_SD3_CMD__USDHC3_CMD   | MUX_PAD_CTRL(USDHC_PAD_CTRL),
	MX6_PAD_SD3_DAT0__USDHC3_DAT0 | MUX_PAD_CTRL(USDHC_PAD_CTRL),
	MX6_PAD_SD3_DAT1__USDHC3_DAT1 | MUX_PAD_CTRL(USDHC_PAD_CTRL),
	MX6_PAD_SD3_DAT2__USDHC3_DAT2 | MUX_PAD_CTRL(USDHC_PAD_CTRL),
	MX6_PAD_SD3_DAT3__USDHC3_DAT3 | MUX_PAD_CTRL(USDHC_PAD_CTRL),
	MX6_PAD_SD3_DAT4__USDHC3_DAT4 | MUX_PAD_CTRL(USDHC_PAD_CTRL),
	MX6_PAD_SD3_DAT5__USDHC3_DAT5 | MUX_PAD_CTRL(USDHC_PAD_CTRL),
	MX6_PAD_SD3_DAT6__USDHC3_DAT6 | MUX_PAD_CTRL(USDHC_PAD_CTRL),
	MX6_PAD_SD3_DAT7__USDHC3_DAT7 | MUX_PAD_CTRL(USDHC_PAD_CTRL),
	MX6_PAD_NANDF_D0__GPIO_2_0    | MUX_PAD_CTRL(NO_PAD_CTRL), /* CD */
};

iomux_v3_cfg_t const usdhc4_pads[] = {
	MX6_PAD_SD4_CLK__USDHC4_CLK   | MUX_PAD_CTRL(USDHC_PAD_CTRL),
	MX6_PAD_SD4_CMD__USDHC4_CMD   | MUX_PAD_CTRL(USDHC_PAD_CTRL),
	MX6_PAD_SD4_DAT0__USDHC4_DAT0 | MUX_PAD_CTRL(USDHC_PAD_CTRL),
	MX6_PAD_SD4_DAT1__USDHC4_DAT1 | MUX_PAD_CTRL(USDHC_PAD_CTRL),
	MX6_PAD_SD4_DAT2__USDHC4_DAT2 | MUX_PAD_CTRL(USDHC_PAD_CTRL),
	MX6_PAD_SD4_DAT3__USDHC4_DAT3 | MUX_PAD_CTRL(USDHC_PAD_CTRL),
	MX6_PAD_SD4_DAT4__USDHC4_DAT4 | MUX_PAD_CTRL(USDHC_PAD_CTRL),
	MX6_PAD_SD4_DAT5__USDHC4_DAT5 | MUX_PAD_CTRL(USDHC_PAD_CTRL),
	MX6_PAD_SD4_DAT6__USDHC4_DAT6 | MUX_PAD_CTRL(USDHC_PAD_CTRL),
	MX6_PAD_SD4_DAT7__USDHC4_DAT7 | MUX_PAD_CTRL(USDHC_PAD_CTRL),
};

static void setup_iomux_uart(void)
{
	imx_iomux_v3_setup_multiple_pads(uart1_pads, ARRAY_SIZE(uart1_pads));
}

#ifdef CONFIG_FSL_ESDHC
struct fsl_esdhc_cfg usdhc_cfg[3] = {
	{USDHC2_BASE_ADDR},
	{USDHC3_BASE_ADDR},
	{USDHC4_BASE_ADDR},
};

#define USDHC2_CD_GPIO	IMX_GPIO_NR(2, 2)
#define USDHC3_CD_GPIO	IMX_GPIO_NR(2, 0)

// GPIO4 Up GPIO5 Down
#define GPIO_VOL_UP     IMX_GPIO_NR(1, 4)
#define GPIO_VOL_DOWN   IMX_GPIO_NR(1, 5)

int board_mmc_getcd(struct mmc *mmc)
{
	struct fsl_esdhc_cfg *cfg = (struct fsl_esdhc_cfg *)mmc->priv;
	int ret = 0;

	switch (cfg->esdhc_base) {
	case USDHC2_BASE_ADDR:
		ret = !gpio_get_value(USDHC2_CD_GPIO);
		break;
	case USDHC3_BASE_ADDR:
		ret = !gpio_get_value(USDHC3_CD_GPIO);
		break;
	case USDHC4_BASE_ADDR:
		ret = 1; /* eMMC/uSDHC4 is always present */
		break;
	}

	return ret;
}

int board_mmc_init(bd_t *bis)
{
	s32 status = 0;
	int i;

	/*
	 * According to the board_mmc_init() the following map is done:
	 * (U-boot device node)    (Physical Port)
	 * mmc0                    SD2
	 * mmc1                    SD3
	 * mmc2                    eMMC
	 */
	for (i = 0; i < CONFIG_SYS_FSL_USDHC_NUM; i++) {
		switch (i) {
		case 0:
			imx_iomux_v3_setup_multiple_pads(
				usdhc2_pads, ARRAY_SIZE(usdhc2_pads));
			gpio_direction_input(USDHC2_CD_GPIO);
			usdhc_cfg[0].sdhc_clk = mxc_get_clock(MXC_ESDHC2_CLK);
			break;
		case 1:
			imx_iomux_v3_setup_multiple_pads(
				usdhc3_pads, ARRAY_SIZE(usdhc3_pads));
			gpio_direction_input(USDHC3_CD_GPIO);
			usdhc_cfg[1].sdhc_clk = mxc_get_clock(MXC_ESDHC3_CLK);
	    usdhc_cfg[1].sdhc_clk = MIN(20000000, usdhc_cfg[1].sdhc_clk);
			break;
		case 2:
			imx_iomux_v3_setup_multiple_pads(
				usdhc4_pads, ARRAY_SIZE(usdhc4_pads));
			usdhc_cfg[2].sdhc_clk = mxc_get_clock(MXC_ESDHC4_CLK);
			break;
		default:
			printf("Warning: you configured more USDHC controllers"
			       "(%d) then supported by the board (%d)\n",
			       i + 1, CONFIG_SYS_FSL_USDHC_NUM);
			return status;
		}


		status |= fsl_esdhc_initialize(bis, &usdhc_cfg[i]);
	}

	return status;
}
#endif

int mx6_rgmii_rework(struct phy_device *phydev)
{
	unsigned short val;

	/* To enable AR8031 ouput a 125MHz clk from CLK_25M */
	phy_write(phydev, MDIO_DEVAD_NONE, 0xd, 0x7);
	phy_write(phydev, MDIO_DEVAD_NONE, 0xe, 0x8016);
	phy_write(phydev, MDIO_DEVAD_NONE, 0xd, 0x4007);

	val = phy_read(phydev, MDIO_DEVAD_NONE, 0xe);
	val &= 0xffe3;
	val |= 0x18;
	phy_write(phydev, MDIO_DEVAD_NONE, 0xe, val);

	/* introduce tx clock delay */
	phy_write(phydev, MDIO_DEVAD_NONE, 0x1d, 0x5);
	val = phy_read(phydev, MDIO_DEVAD_NONE, 0x1e);
	val |= 0x0100;
	phy_write(phydev, MDIO_DEVAD_NONE, 0x1e, val);

	return 0;
}

int board_phy_config(struct phy_device *phydev)
{
	mx6_rgmii_rework(phydev);

	if (phydev->drv->config)
		phydev->drv->config(phydev);

	return 0;
}

int board_eth_init(bd_t *bis)
{
	int ret;

	setup_iomux_enet();

	ret = cpu_eth_init(bis);
	if (ret)
		printf("FEC MXC: %s:failed\n", __func__);

	return 0;
}


#define USBOTG_IN_PAD_CTRL  (PAD_CTL_HYS | PAD_CTL_PUS_100K_DOWN | PAD_CTL_DSE_240ohm | PAD_CTL_SRE_SLOW)
#define USBOTG_OUT_PAD_CTRL (                                      PAD_CTL_DSE_240ohm | PAD_CTL_SRE_SLOW)

static void setup_iomux_usbotg(void)
{
  static const iomux_v3_cfg_t usbotg_pads[] =
  {
    NEW_PAD_CTRL(MX6_PAD_EIM_D22__USBOH3_USBOTG_PWR,
        USBOTG_OUT_PAD_CTRL),
    NEW_PAD_CTRL(MX6_PAD_EIM_D21__USBOH3_USBOTG_OC,
        USBOTG_IN_PAD_CTRL),
  };

  /* Set up pins for USBOTG. */
  imx_iomux_v3_setup_multiple_pads(usbotg_pads, ARRAY_SIZE(usbotg_pads));
}


int board_early_init_f(void)
{
  setup_iomux_usbotg();
	setup_iomux_uart();

	return 0;
}

int board_init(void)
{
	/* address of boot parameters */
	gd->bd->bi_boot_params = PHYS_SDRAM + 0x100;

  return 0;
}

#define PFUZE100_SW1ABVOL 0x20
#define PFUZE100_SW1CVOL  0x2e
#define PFUZE100_SW2VOL   0x35
#define PFUZE100_SW3AVOL  0x3c
#define PFUZE100_SW3BVOL  0x43
#define PFUZE100_SW4VOL   0x4a

struct PFUZE100VoltageRange_s
{
  int Min;
  int Max;
  int Inc;
};

struct PFUZE100Voltages_s
{
  char *Name;
  int Reg;
  int CorValue;
  struct PFUZE100VoltageRange_s LowRange;
  struct PFUZE100VoltageRange_s HighRange;
};

#define PFUZE100_SW_REG(_Name, _RegCorrValue, _RegVolt, _MinLow, _MaxLow, _IncrementLow, _MinHigh, _MaxHigh, _IncrementHigh)            \
  {                                                                                        \
    .Name = #_Name, .Reg = _RegVolt, .CorValue = _RegCorrValue,                            \
    .LowRange.Min = _MinLow, .LowRange.Max = _MaxLow, .LowRange.Inc = _IncrementLow,       \
    .HighRange.Min = _MinHigh, .HighRange.Max = _MaxHigh, .HighRange.Inc = _IncrementHigh, \
  }

static struct PFUZE100Voltages_s PFUZE100Voltages[] =
{
  PFUZE100_SW_REG(SW1AB, 0x00, PFUZE100_SW1ABVOL,  300000, 1075000, 25000, 1100000, 1875000, 25000),
  PFUZE100_SW_REG(SW1C,  0x00, PFUZE100_SW1CVOL,   300000, 1075000, 25000, 1100000, 1875000, 25000),
  PFUZE100_SW_REG(SW2,   0x00, PFUZE100_SW2VOL,    400000, 1975000, 25000, 800000,  3300000, 50000),
  PFUZE100_SW_REG(SW3A,  0x00, PFUZE100_SW3AVOL,   400000, 1975000, 25000, 800000,  3300000, 50000),
  PFUZE100_SW_REG(SW3B,  0x00, PFUZE100_SW3BVOL,   400000, 1975000, 25000, 800000,  3300000, 50000),
  PFUZE100_SW_REG(SW4,   0x6A, PFUZE100_SW4VOL,    400000, 1975000, 25000, 800000,  3300000, 50000),
};


int pfuze100_init(void)
{
  setup_i2c(1, CONFIG_SYS_I2C_SPEED, 0x7f, &i2c_pad_info1);

  unsigned char PfuzeRegVal;

  // Setzen der alternativspannungen wenn der Defaultwert nicht stimmt dauert ca 350ms
  int i;
  for (i = 0; i < sizeof(PFUZE100Voltages)/sizeof(PFUZE100Voltages[0]); i++)
  {
    struct PFUZE100Voltages_s VoltageDef = PFUZE100Voltages[i];
    if (i2c_read(CONFIG_SYS_I2C_PMIC_PFUZE100_ADDR, VoltageDef.Reg, 1, (void *) &PfuzeRegVal, sizeof(PfuzeRegVal)) == 0)
    {
      struct PFUZE100VoltageRange_s Range = VoltageDef.LowRange;
      if (PfuzeRegVal & 0x40) { Range = VoltageDef.HighRange; }
      printf("PMIC: Default %s voltage %dmV (%02X)", VoltageDef.Name, Range.Min + Range.Inc*(PfuzeRegVal & 0x3F), PfuzeRegVal);
      if (VoltageDef.CorValue && PfuzeRegVal != VoltageDef.CorValue)
      {
        PfuzeRegVal = VoltageDef.CorValue;
        if (i2c_write(CONFIG_SYS_I2C_PMIC_PFUZE100_ADDR, 0x4A, 1, (void *) &PfuzeRegVal, sizeof(PfuzeRegVal)) == 0)
        {
        }
        if (PfuzeRegVal & 0x40) { Range = VoltageDef.HighRange; }
        printf(" changed value to %s: %dmV", VoltageDef.Name, Range.Min + Range.Inc*(PfuzeRegVal & 0x3F));
      }
      printf("\n");
    }
  }

  printf("PMIC: Register dump:\n");
  for (i = 0; i < 128; i++)
  {
    if (i2c_read(CONFIG_SYS_I2C_PMIC_PFUZE100_ADDR, i, 1, (void *) &PfuzeRegVal, sizeof(PfuzeRegVal)) == 0)
    {
      printf("%02X ", PfuzeRegVal);
      if (((i + 1) % 16) == 0) printf("\n");
    }
  }
  printf("\n");
  return 0;
}




int misc_init_r(void)
{
  pfuze100_init();

  gpio_direction_input(GPIO_VOL_UP);
  gpio_direction_input(GPIO_VOL_DOWN);

  if (!gpio_get_value(GPIO_VOL_UP))
    printf("Volume-UP Button Pressed\n");
  if (!gpio_get_value(GPIO_VOL_DOWN))
    printf("Volume-DOWN Button Pressed\n");

  if (!gpio_get_value(GPIO_VOL_DOWN))
  {
    setenv("bootdelay", "10");
  }
  else
    setenv("bootdelay", "0");

  return 0;
}

#ifdef CONFIG_CMD_BMODE
static const struct boot_mode board_boot_modes[] = {
	/* 4 bit bus width */
	{"sd2",	 MAKE_CFGVAL(0x40, 0x28, 0x00, 0x00)},
	{"sd3",	 MAKE_CFGVAL(0x40, 0x30, 0x00, 0x00)},
	/* 8 bit bus width */
	{"emmc", MAKE_CFGVAL(0x40, 0x38, 0x00, 0x00)},
	{NULL,	 0},
};
#endif


int last_stage_init(void)
{
#ifdef CONFIG_UCP
  if (!gpio_get_value(GPIO_VOL_UP))
  {
    printf("%s -> UCP\n", __func__);
    setenv("ipaddr", "192.168.0.1");
    setenv("serverip", "192.168.0.1");
    NetStartAgain();
    NetLoop(UCP);
  }
#endif

  return 0;
}

int board_late_init(void)
{
#ifdef CONFIG_CMD_BMODE
	add_board_boot_modes(board_boot_modes);
#endif

	return 0;
}

#pragma pack(1)

typedef unsigned char __uint8_t;
typedef unsigned short __uint16_t;
typedef unsigned int __uint32_t;

typedef struct __attribute__ ((packed)) FileUpgHeader_s
{
  __uint32_t HdrMagic;
  __uint32_t HdrSize;
  __uint8_t Name[64];
  __uint8_t Data[0];
} FileUpgHeader_t;

typedef struct __attribute__ ((packed)) UpgHeader_s
{
  __uint32_t HdrMagic;
  __uint32_t HdrSize;
  __uint8_t Name[64];
  __uint8_t SoftwareVersion[64];
  __uint8_t HardwareVersion[64];
  __uint32_t UpgradeSize;
  __uint8_t Md5Sum[16];
  __uint8_t Sha1Sum[16];
  __uint32_t SignType;
  __uint8_t SignatureData[128];
  FileUpgHeader_t UpgradeFiles[0];
} UpgHeader_t;

#pragma pack()


static int init_mmc_for_env(struct mmc *mmc)
{
  if (!mmc) {
    puts("No MMC card found\n");
    return -1;
  }

  if (mmc_init(mmc)) {
    puts("MMC init failed\n");
    return -1;
  }

  return 0;
}

#ifdef CONFIG_UCP
void UCPExecuteFirmwareUpgrade(void *Addr, size_t Len)
{
  printf("UCPExecuteFirmwareUpgrade -> %p %d\n", Addr, Len);
  UpgHeader_t *UpgData = (UpgHeader_t *) Addr;
  printf("Upgrade Magic: %08X\n", UpgData->HdrMagic);
  printf("Upgrade Size: %08X\n", UpgData->UpgradeSize);
  printf("Upgrade Name: %s\n", UpgData->Name);

  __uint8_t Sha1Sum[SHA1_SUM_LEN];

  const unsigned char *UpgradeData = Addr + sizeof(UpgHeader_t);
  int UpgradeDataLen = Len - sizeof(UpgHeader_t);
  sha1_csum(UpgradeData, UpgradeDataLen, Sha1Sum);

#if 0
  int i;
  printf("SHA1 Calc: ");
  for (i = 0; i < SHA1_SUM_LEN; i++)
  {
    printf("%02X", Sha1Sum[i]);
  }
  printf("\n");

  printf("SHA1 Header: ");
  for (i = 0; i < SHA1_SUM_LEN; i++)
  {
    printf("%02X", UpgData->Sha1Sum[i]);
  }
  printf("\n");
#endif

  if (memcmp(Sha1Sum, UpgData->Sha1Sum, SHA1_SUM_LEN) == 0)
  {
    printf("SHA1 OK!\n");
  } else
  {
    printf("SHA1 Failed!\n");
  }

  int Offset = 1024;
  if (strcmp("UBootIMX6", (char *) UpgData->Name) == 0)
  {
    printf("UBoot upgrade\n");
    Offset = 1024;
  } else
  {
    printf("Unknown Upgrade Type...\n");
    return;
  }

#ifdef CONFIG_ENV_IS_IN_SPI_FLASH

#endif

#ifdef CONFIG_ENV_IS_IN_MMC
  struct mmc *MMCDev = find_mmc_device(CONFIG_SYS_MMC_ENV_DEV);
  if (init_mmc_for_env(MMCDev)) return;

  const unsigned char *Buffer = UpgradeData;
  int Size = UpgradeDataLen;
  uint BlkStart, BlkCount, n;

  BlkStart = ALIGN(Offset, MMCDev->write_bl_len) / MMCDev->write_bl_len;
  BlkCount = ALIGN(Size, MMCDev->write_bl_len) / MMCDev->write_bl_len;
  n = MMCDev->block_dev.block_write(CONFIG_SYS_MMC_ENV_DEV, BlkStart, BlkCount, (u_char *) Buffer);

  if (n != BlkCount)
  {
    printf("Write to MMC Failed!\n");
  } else
  {
    printf("Write to MMC OK!\n");
    do_reset(NULL, 0, 0, NULL);
  }
#endif
}
#endif

int checkboard(void)
{
	puts("Board: MX6Q-SabreSD\n");

	return 0;
}
