/*
 * ddrcal.c
 *
 *  Created on: 05.03.2014
 *      Author: thomas
 */

#include <common.h>
#include <asm/io.h>
#include <asm/arch/clock.h>
#include <asm/arch/imx-regs.h>
#include <asm/arch/iomux.h>
#include <asm/arch/mx6-pins.h>
#include <asm/arch/mx6-ddr.h>
#include <asm/errno.h>
#include <asm/gpio.h>
#include <asm/imx-common/iomux-v3.h>
#include <asm/imx-common/boot_mode.h>

#include "mx6smodas.h"

#ifndef reg32_write
#define reg32_write writel
#define reg32_read readl
#endif

void reg32setbit(int Reg, int Bit)
{
  int NewValue = (reg32_read(Reg) | (1 << Bit));
  reg32_write(Reg, NewValue);
}

void reg32clrbit(int Reg, int Bit)
{
  int NewValue = (reg32_read(Reg) & (~(1 << Bit)));
  reg32_write(Reg, NewValue);
}

#define MMDC_P0_IPS_BASE_ADDR 0x021b0000
#define MMDC_P1_IPS_BASE_ADDR 0x021b4000

#define MDCTL_OFFSET      0x0000
#define MDPDC_OFFSET      0x0004
#define MAPSR_OFFSET      0x0404
#define MDMISC_OFFSET     0x0018
#define MDREF_OFFSET      0x0020
#define MDSCR_OFFSET      0x001C
#define MPPDCMPR1_OFFSET  0x088C
#define MPSWDAR_OFFSET    0x0894
#define MPRDDLCTL_OFFSET  0x088C
#define MPMUR_OFFSET      0x0848
#define MPDGCTRL0_OFFSET  0x083C
#define MPDGCTRL1_OFFSET  0x0840
#define MPDGHWST0_OFFSET  0x087C
#define MPDGHWST1_OFFSET  0x0880
#define MPDGHWST2_OFFSET  0x0884
#define MPDGHWST3_OFFSET  0x0888
#define MPRDDLHWCTL_OFFSET 0x0860
#define MPWRDLCTL_OFFSET 0x0850
#define MPWRDLHWCTL_OFFSET 0x0864

#define IOMUXC_SW_PAD_CTL_PAD_DRAM_SDQS0 MX6_IOM_DRAM_SDQS0
#define IOMUXC_SW_PAD_CTL_PAD_DRAM_SDQS1 MX6_IOM_DRAM_SDQS1
#define IOMUXC_SW_PAD_CTL_PAD_DRAM_SDQS2 MX6_IOM_DRAM_SDQS2
#define IOMUXC_SW_PAD_CTL_PAD_DRAM_SDQS3 MX6_IOM_DRAM_SDQS3
#define IOMUXC_SW_PAD_CTL_PAD_DRAM_SDQS4 MX6_IOM_DRAM_SDQS4
#define IOMUXC_SW_PAD_CTL_PAD_DRAM_SDQS5 MX6_IOM_DRAM_SDQS5
#define IOMUXC_SW_PAD_CTL_PAD_DRAM_SDQS6 MX6_IOM_DRAM_SDQS6
#define IOMUXC_SW_PAD_CTL_PAD_DRAM_SDQS7 MX6_IOM_DRAM_SDQS7

#if 0
int write_level_calib(int ddr_mr1)
{
  int errorcount = 0;
  int temp1, temp2, temp3;
  // disable DDR logic power down timer:
  reg32_write(MX6_MMDC_P0_MDPDC, reg32_read(MX6_MMDC_P0_MDPDC) & 0xffff00ff);
  // disable Adopt power down timer:
  reg32_write(MX6_MMDC_P0_MAPSR, reg32_read(MX6_MMDC_P0_MAPSR) | 0x1);
  printf("Start write leveling calibration \n");
  // disable auto refresh and ZQ calibration
  // before proceeding with Write Leveling calibration
  temp1 = reg32_read(MX6_MMDC_P0_MDREF);
  reg32_write(MX6_MMDC_P0_MDREF, 0x0000C000);
  temp2 = reg32_read(MX6_MMDC_P0_MPZQHWCTRL);
  reg32_write(MX6_MMDC_P0_MPZQHWCTRL, temp2 & ~(0x3));
  // Configure the external DDR device to enter write leveling mode
  // through Load Mode Register command
  // Register setting:
  // Bits[31:16] MR1 value (0x0080 write leveling enable)
  // Bit[9] set WL_EN to enable MMDC DQS output
  // Bits[6:4] set CMD bits for Load Mode Register programming
  // Bits[2:0] set CMD_BA to 0x1 for DDR MR1 programming
  reg32_write(MX6_MMDC_P0_MDSCR, 0x00808231);
  // Activate automatic calibration by setting MPWLGCR[HW_WL_EN]
  reg32_write(MX6_MMDC_P0_MPWLGCR, 0x00000001);
  // Upon completion of this process the MMDC de-asserts the MPWLGCR[HW_WL_EN]
  while (reg32_read(MX6_MMDC_P0_MPWLGCR) & 0x00000001);
  // check for any errors: for x32, check only PHY0
  if ((reg32_read(MX6_MMDC_P0_MPWLGCR) & 0x00000F00) )
  {
    errorcount++;
  }
  printf("Write leveling calibration completed\n");
  // User should issue MRS command to exit write leveling mode
  // through Load Mode Register
  // Register setting:
  // Bits[31:16] MR1 value "ddr_mr1" value from initialization
  // Bit[9] clear WL_EN to disable MMDC DQS output
  // Bits[6:4] set CMD bits for Load Mode Register programming
  // Bits[2:0] set CMD_BA to 0x1 for DDR MR1 programming
  reg32_write(MX6_MMDC_P0_MDSCR,((ddr_mr1 << 16)+0x8031));
  // re-enable to auto refresh and zq cal
  reg32_write(MX6_MMDC_P0_MDREF, temp1);
  reg32_write(MX6_MMDC_P0_MPZQHWCTRL, temp2);
  printf("MMDC_MPWLDECTRL0 after write level cal: 0x%08X\n", reg32_read(MX6_MMDC_P0_MPWLDECTRL0));
  printf("MMDC_MPWLDECTRL1 after write level cal: 0x%08X\n", reg32_read(MX6_MMDC_P0_MPWLDECTRL1));
  // enable DDR logic power down timer:
  reg32_write(MX6_MMDC_P0_MDPDC, reg32_read(MX6_MMDC_P0_MDPDC) | 0x00005500);
  // enable Adopt power down timer:
  reg32_write(MX6_MMDC_P0_MAPSR, reg32_read(MX6_MMDC_P0_MAPSR) & 0xfffffff7);
  reg32_write(MX6_MMDC_P0_MDSCR, 0); //clear CON_REQ
  return errorcount;
}
#endif

int modify_dg_result(int reg_st0, int reg_st1, int reg_ctrl)
{
  // DQS gating absolute offset should be modified from reflecting (HW_DG_LOWx +  HW_DG_UPx)/2
  // to reflecting (HW_DG_UPx - 0x80)
  int dg_tmp_val0,dg_tmp_val1, dg_tmp_val2;
  int dg_dl_abs_offset0, dg_dl_abs_offset1;
  int dg_hc_del0, dg_hc_del1;
  dg_tmp_val0 = ((reg32_read(reg_st0) & 0x07ff0000) >>16) - 0xc0;
  dg_tmp_val1 = ((reg32_read(reg_st1) & 0x07ff0000) >>16) - 0xc0;
  dg_dl_abs_offset0 = dg_tmp_val0 & 0x7f;
  dg_hc_del0 = (dg_tmp_val0 & 0x780) << 1;
  dg_dl_abs_offset1 = dg_tmp_val1 & 0x7f;
  dg_hc_del1 = (dg_tmp_val1 & 0x780) << 1;
  dg_tmp_val2 = dg_dl_abs_offset0 + dg_hc_del0 + ((dg_dl_abs_offset1 +
      dg_hc_del1) << 16);
  reg32_write((reg_ctrl),
              reg32_read((reg_ctrl)) & 0xf0000000);
  reg32_write((reg_ctrl),
              reg32_read((reg_ctrl)) & 0xf0000000);
  reg32_write((reg_ctrl),
              reg32_read((reg_ctrl)) | dg_tmp_val2);
  return 0;
}

static inline int setgpio(unsigned gpio, int value)
{
  if (value)
    writel(readl(GPIO6_BASE_ADDR) | (1 << (gpio & 0x1f)), GPIO6_BASE_ADDR);
  else
    writel(readl(GPIO6_BASE_ADDR) & ~(1 << (gpio & 0x1f)), GPIO6_BASE_ADDR);
  return 0;
}

int ddr3cal(void)
{
  int errorcount = 0;
  int temp1;
  int  __attribute__((__unused__)) g_error_read_dqs_gating;
  int  __attribute__((__unused__)) g_error_read_cal;
  int  __attribute__((__unused__)) g_error_write_cal;
  int data_bus_size;
  int temp_ref;
  int cs0_enable = 0;
  int cs1_enable = 0;
  int cs0_enable_initial = 0;
  int cs1_enable_initial = 0;
  // int PDDWord = 0x55aaaa55; // original values, these work, but can get slightly better below
  int PDDWord = 0x00FFFF00; // best so far, place into MPPDCMPR1

  setgpio(LED_POWER_R, 1);

  // check to see which chip selects are enabled
  cs0_enable_initial = (reg32_read(MMDC_P0_IPS_BASE_ADDR + MDCTL_OFFSET) & 0x80000000) >> 31;
  cs1_enable_initial = (reg32_read(MMDC_P0_IPS_BASE_ADDR + MDCTL_OFFSET) & 0x40000000) >> 30;

  setgpio(LED_SER_R, 1);

  // printf("DDR: Power Down... CS0:%d CS1:%d\n", cs0_enable_initial, cs1_enable_initial);

  return 444;

  // disable DDR logic power down timer:
  // printf("DDR: MDPDC: %08X -> %08X\n", reg32_read((MMDC_P0_IPS_BASE_ADDR + MDPDC_OFFSET)), reg32_read((MMDC_P0_IPS_BASE_ADDR + MDPDC_OFFSET)) & 0xffff00ff);
  reg32_write((MMDC_P0_IPS_BASE_ADDR + MDPDC_OFFSET), reg32_read((MMDC_P0_IPS_BASE_ADDR + MDPDC_OFFSET)) & 0xffff00ff);

  setgpio(LED_SER_G, 1);



  // printf("DDR: disable Adopt power down timer\n");
  // disable Adopt power down timer:
  reg32_write((MMDC_P0_IPS_BASE_ADDR + MAPSR_OFFSET), reg32_read((MMDC_P0_IPS_BASE_ADDR + MAPSR_OFFSET)) | 0x1);

  // printf("DDR: set the device ODT during read\n");

  // set the device ODT during read:
  // reg32_write((MMDC_P0_IPS_BASE_ADDR + MPODTCTRL_OFFSET),
  // reg32_read((MMDC_P0_IPS_BASE_ADDR + MPODTCTRL_OFFSET)) | 0x8);
  //set DQS pull ups
  reg32_write(IOMUXC_SW_PAD_CTL_PAD_DRAM_SDQS0,
              reg32_read(IOMUXC_SW_PAD_CTL_PAD_DRAM_SDQS0) | 0x7000);
  reg32_write(IOMUXC_SW_PAD_CTL_PAD_DRAM_SDQS1,
              reg32_read(IOMUXC_SW_PAD_CTL_PAD_DRAM_SDQS1) | 0x7000);
#if 0
  reg32_write(IOMUXC_SW_PAD_CTL_PAD_DRAM_SDQS2,
              reg32_read(IOMUXC_SW_PAD_CTL_PAD_DRAM_SDQS2) | 0x7000);
  reg32_write(IOMUXC_SW_PAD_CTL_PAD_DRAM_SDQS3,
              reg32_read(IOMUXC_SW_PAD_CTL_PAD_DRAM_SDQS3) | 0x7000);
  reg32_write(IOMUXC_SW_PAD_CTL_PAD_DRAM_SDQS4,
              reg32_read(IOMUXC_SW_PAD_CTL_PAD_DRAM_SDQS4) | 0x7000);
  reg32_write(IOMUXC_SW_PAD_CTL_PAD_DRAM_SDQS5,
              reg32_read(IOMUXC_SW_PAD_CTL_PAD_DRAM_SDQS5) | 0x7000);
  reg32_write(IOMUXC_SW_PAD_CTL_PAD_DRAM_SDQS6,
              reg32_read(IOMUXC_SW_PAD_CTL_PAD_DRAM_SDQS6) | 0x7000);
  reg32_write(IOMUXC_SW_PAD_CTL_PAD_DRAM_SDQS7,
              reg32_read(IOMUXC_SW_PAD_CTL_PAD_DRAM_SDQS7) | 0x7000);
#endif
  temp1 = reg32_read(MMDC_P0_IPS_BASE_ADDR + MDMISC_OFFSET);
  reg32setbit((MMDC_P0_IPS_BASE_ADDR + MDMISC_OFFSET), 6); //set RALAT to max
  reg32setbit((MMDC_P0_IPS_BASE_ADDR + MDMISC_OFFSET), 7);
  reg32setbit((MMDC_P0_IPS_BASE_ADDR + MDMISC_OFFSET), 8);
  reg32setbit((MMDC_P0_IPS_BASE_ADDR + MDMISC_OFFSET), 16); //set WALAT to max
  reg32setbit((MMDC_P0_IPS_BASE_ADDR + MDMISC_OFFSET), 17);


  // disable auto refresh
  // before proceeding with calibration
  temp_ref = reg32_read(MMDC_P0_IPS_BASE_ADDR + MDREF_OFFSET);
  reg32_write((MMDC_P0_IPS_BASE_ADDR + MDREF_OFFSET), 0x0000C000);
  // per the ref manual, issue one refresh cycle MDSCR[CMD]= 0x2, this also sets the CON_REQ bit.
  if (cs0_enable_initial)
    reg32_write((MMDC_P0_IPS_BASE_ADDR + MDSCR_OFFSET), 0x00008020);
  if (cs1_enable_initial)
    reg32_write((MMDC_P0_IPS_BASE_ADDR + MDSCR_OFFSET), 0x00008028);
  // check MDMISC register CALIB_PER_CS to see which CS calibration is targeted to (under normal
  // cases, it should be cleared as this is the default value, indicating calibration is directed to CS0).
  // Disable the other chip select not being target for calibration to avoid any potential issues
  // This will get re-enabled at end of calibration
  if ((reg32_read(MMDC_P0_IPS_BASE_ADDR + MDMISC_OFFSET) & 0x00100000) == 0)
  {
    reg32clrbit((MMDC_P0_IPS_BASE_ADDR + MDCTL_OFFSET), 30); // clear SDE_1
  } else
  {
    reg32clrbit((MMDC_P0_IPS_BASE_ADDR + MDCTL_OFFSET), 31); // clear SDE_0
  }

  // check to see which chip selects are now enabled for the remainder of the calibration
  cs0_enable = (reg32_read(MMDC_P0_IPS_BASE_ADDR + MDCTL_OFFSET) & 0x80000000) >> 31;
  cs1_enable = (reg32_read(MMDC_P0_IPS_BASE_ADDR + MDCTL_OFFSET) & 0x40000000) >> 30;
  // check to see what is the data bus size:
  data_bus_size= (reg32_read(MMDC_P0_IPS_BASE_ADDR + MDCTL_OFFSET) & 0x30000) >> 16;
  // Issue the Precharge-All command to the DDR device for both chip selects
  // Note, CON_REQ bit should also remain set
  // If only using one chip select, then precharge only the desired chip select
  if (cs0_enable == 1)
    reg32_write((MMDC_P0_IPS_BASE_ADDR + MDSCR_OFFSET), 0x04008050); // CS0
  if (cs1_enable == 1)
    reg32_write((MMDC_P0_IPS_BASE_ADDR + MDSCR_OFFSET), 0x04008058); // CS1
  // Write the pre-defined value into MPPDCMPR1
  reg32_write((MMDC_P0_IPS_BASE_ADDR + MPPDCMPR1_OFFSET), PDDWord);
  // Issue a write access to the external DDR device by setting the bit SW_DUMMY_WR (bit 0)
  // in the MPSWDAR0 and then poll this bit until it clears to indicate completion of the
  // write access.
  reg32setbit((MMDC_P0_IPS_BASE_ADDR + MPSWDAR_OFFSET), 0);
  while (reg32_read((MMDC_P0_IPS_BASE_ADDR + MPSWDAR_OFFSET)) & 0x00000001);
  // Set the RD_DL_ABS_OFFSET# bits to their default values (will be calibrated later in
  // the read delay-line calibration)
  // Both PHYs for x64 configuration, if x32, do only PHY0
  reg32_write((MMDC_P0_IPS_BASE_ADDR + MPRDDLCTL_OFFSET), 0x40404040);
  if (data_bus_size == 0x2)
  {
    reg32_write((MMDC_P1_IPS_BASE_ADDR + MPRDDLCTL_OFFSET), 0x40404040);
  }

  setgpio(LED_LAN_R, 1);

  //Force a measurment, for previous delay setup to take effect:
  reg32_write((MMDC_P0_IPS_BASE_ADDR + MPMUR_OFFSET), 0x800);
  if (data_bus_size == 0x2)
  {
    reg32_write((MMDC_P1_IPS_BASE_ADDR + MPMUR_OFFSET), 0x800);
  }

  //@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
  // Read DQS Gating calibration
  //@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
  // printf("Starting DQS gating calibration...\n");
  // Reset the read data FIFOs (two resets); only need to issue reset to PHY0 since in x64
  // mode, the reset will also go to PHY1
  // read data FIFOs reset1
  reg32_write((MMDC_P0_IPS_BASE_ADDR + MPDGCTRL0_OFFSET),
              reg32_read((MMDC_P0_IPS_BASE_ADDR + MPDGCTRL0_OFFSET)) | 0x80000000);
  while (reg32_read((MMDC_P0_IPS_BASE_ADDR + MPDGCTRL0_OFFSET)) & 0x80000000);
  // read data FIFOs reset2
  reg32_write((MMDC_P0_IPS_BASE_ADDR + MPDGCTRL0_OFFSET),
              reg32_read((MMDC_P0_IPS_BASE_ADDR + MPDGCTRL0_OFFSET)) | 0x80000000);
  while (reg32_read((MMDC_P0_IPS_BASE_ADDR + MPDGCTRL0_OFFSET)) & 0x80000000);
  // Start the automatic read DQS gating calibration process by asserting MPDGCTRL0[HW_DG_EN]
  // and MPDGCTRL0[DG_CMP_CYC] and then poll MPDGCTRL0[HW_DG_EN]] until this bit clears to
  // indicate completion.
  // Also, ensure that MPDGCTRL0[HW_DG_ERR] is clear to indicate no errors were seen during
  // calibration.
  // Set bit 30: chooses option to wait 32 cycles instead of 16 before comparing read data
  reg32setbit((MMDC_P0_IPS_BASE_ADDR + MPDGCTRL0_OFFSET), 30);
  // Set bit 28 to start automatic read DQS gating calibration
  reg32setbit((MMDC_P0_IPS_BASE_ADDR + MPDGCTRL0_OFFSET), 28);
  // Poll for completion
  // MPDGCTRL0[HW_DG_EN] should be 0
  while (reg32_read((MMDC_P0_IPS_BASE_ADDR + MPDGCTRL0_OFFSET)) & 0x10000000);
  // Check to see if any errors were encountered during calibration
  // (check MPDGCTRL0[HW_DG_ERR])
  // check both PHYs for x64 configuration, if x32, check only PHY0
  if (data_bus_size == 0x2)
  {
    if ((reg32_read(MMDC_P0_IPS_BASE_ADDR + MPDGCTRL0_OFFSET) & 0x00001000) ||
        (reg32_read(MMDC_P1_IPS_BASE_ADDR + MPDGCTRL0_OFFSET) & 0x00001000))
    {
      errorcount++; // increment the errorcount variable
      g_error_read_dqs_gating = 1; // set the g_error_read_dqs_gating
    }
  } else
  {
    if (reg32_read(MMDC_P0_IPS_BASE_ADDR + MPDGCTRL0_OFFSET) & 0x00001000)
    {
      errorcount++; // increment the errorcount variable
      g_error_read_dqs_gating = 1; // set the g_error_read_dqs_gating
    }
  }
  // DQS gating absolute offset should be modified from reflecting (HW_DG_LOWx + HW_DG_UPx)/2
  // to reflecting (HW_DG_UPx - 0x80)
  modify_dg_result(MMDC_P0_IPS_BASE_ADDR + MPDGHWST0_OFFSET,
                   MMDC_P0_IPS_BASE_ADDR + MPDGHWST1_OFFSET,
                   MMDC_P0_IPS_BASE_ADDR + MPDGCTRL0_OFFSET);
  modify_dg_result(MMDC_P0_IPS_BASE_ADDR + MPDGHWST2_OFFSET,
                   MMDC_P0_IPS_BASE_ADDR + MPDGHWST3_OFFSET,
                   MMDC_P0_IPS_BASE_ADDR + MPDGCTRL1_OFFSET);
  if (data_bus_size == 0x2)
  {
    modify_dg_result((MMDC_P1_IPS_BASE_ADDR + MPDGHWST0_OFFSET),
                     (MMDC_P1_IPS_BASE_ADDR + MPDGHWST1_OFFSET),
                     (MMDC_P1_IPS_BASE_ADDR + MPDGCTRL0_OFFSET));
    modify_dg_result((MMDC_P1_IPS_BASE_ADDR + MPDGHWST2_OFFSET),
                     (MMDC_P1_IPS_BASE_ADDR + MPDGHWST3_OFFSET),
                     (MMDC_P1_IPS_BASE_ADDR + MPDGCTRL1_OFFSET));
  }
  // printf("DQS gating calibration completed.\n");
  //@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
  // Read delay Calibration
  //@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
  // printf("Starting read calibration...\n");
  // Reset the read data FIFOs (two resets); only need to issue reset to PHY0 since in x64
  // mode, the reset will also go to PHY1
  // read data FIFOs reset1
  reg32_write((MMDC_P0_IPS_BASE_ADDR + MPDGCTRL0_OFFSET),
              reg32_read((MMDC_P0_IPS_BASE_ADDR + MPDGCTRL0_OFFSET)) | 0x80000000);
  while (reg32_read((MMDC_P0_IPS_BASE_ADDR + MPDGCTRL0_OFFSET)) & 0x80000000);
  // read data FIFOs reset2
  reg32_write((MMDC_P0_IPS_BASE_ADDR + MPDGCTRL0_OFFSET),
              reg32_read((MMDC_P0_IPS_BASE_ADDR + MPDGCTRL0_OFFSET)) | 0x80000000);
  while (reg32_read((MMDC_P0_IPS_BASE_ADDR + MPDGCTRL0_OFFSET)) & 0x80000000);
  // Issue the Precharge-All command to the DDR device for both chip selects
  // If only using one chip select, then precharge only the desired chip select
  if (cs0_enable == 1)
    reg32_write((MMDC_P0_IPS_BASE_ADDR + MDSCR_OFFSET), 0x04008050); // CS0
  if (cs1_enable == 1)
    reg32_write((MMDC_P0_IPS_BASE_ADDR + MDSCR_OFFSET), 0x04008058); // CS1
  // Read delay-line calibration
  // Start the automatic read calibration process by asserting MPRDDLHWCTL[HW_RD_DL_EN]
  reg32_write((MMDC_P0_IPS_BASE_ADDR + MPRDDLHWCTL_OFFSET), 0x00000030);
  // poll for completion
  // MMDC indicates that the write data calibration had finished by setting
  // MPRDDLHWCTL[HW_RD_DL_EN] = 0
  // Also, ensure that no error bits were set
  while (reg32_read((MMDC_P0_IPS_BASE_ADDR + MPRDDLHWCTL_OFFSET)) & 0x00000010) ;
  // check both PHYs for x64 configuration, if x32, check only PHY0
  if (data_bus_size == 0x2)
  {
    if ((reg32_read(MMDC_P0_IPS_BASE_ADDR + MPRDDLHWCTL_OFFSET) & 0x0000000f) ||
        (reg32_read(MMDC_P1_IPS_BASE_ADDR + MPRDDLHWCTL_OFFSET) & 0x0000000f))
    {
      errorcount++; // increment the errorcount variable
      g_error_read_cal = 1; // set the g_error_read_cal
    }
  } else
  {
    if (reg32_read(MMDC_P0_IPS_BASE_ADDR + MPRDDLHWCTL_OFFSET) & 0x0000000f) {
      errorcount++; // increment the errorcount variable
      g_error_read_cal = 1; // set the g_error_read_cal
    }
  }
  // printf("Read calibration completed\n");
  //@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
  // Write delay Calibration
  //@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
  // printf("Starting write calibration...\n");
  // Reset the read data FIFOs (two resets); only need to issue reset to PHY0 since in x64
  // mode, the reset will also go to PHY1
  // read data FIFOs reset1
  reg32_write((MMDC_P0_IPS_BASE_ADDR + MPDGCTRL0_OFFSET),
              reg32_read((MMDC_P0_IPS_BASE_ADDR + MPDGCTRL0_OFFSET)) | 0x80000000);
  while (reg32_read((MMDC_P0_IPS_BASE_ADDR + MPDGCTRL0_OFFSET)) & 0x80000000)
    ;
  // read data FIFOs reset2
  reg32_write((MMDC_P0_IPS_BASE_ADDR + MPDGCTRL0_OFFSET),
              reg32_read((MMDC_P0_IPS_BASE_ADDR + MPDGCTRL0_OFFSET)) | 0x80000000);
  while (reg32_read((MMDC_P0_IPS_BASE_ADDR + MPDGCTRL0_OFFSET)) & 0x80000000)
    ;
  // Issue the Precharge-All command to the DDR device for both chip selects
  // If only using one chip select, then precharge only the desired chip select
  if (cs0_enable == 1)
    reg32_write((MMDC_P0_IPS_BASE_ADDR + MDSCR_OFFSET), 0x04008050); // CS0
  if (cs1_enable == 1)
    reg32_write((MMDC_P0_IPS_BASE_ADDR + MDSCR_OFFSET), 0x04008058); // CS1
  // Set the WR_DL_ABS_OFFSET# bits to their default values
  // Both PHYs for x64 configuration, if x32, do only PHY0
  reg32_write((MMDC_P0_IPS_BASE_ADDR + MPWRDLCTL_OFFSET), 0x40404040);
  if (data_bus_size == 0x2)
  {
    reg32_write((MMDC_P1_IPS_BASE_ADDR + MPWRDLCTL_OFFSET), 0x40404040);
  }
  //Force a measurment, for previous delay setup to effect:
  reg32_write((MMDC_P0_IPS_BASE_ADDR + MPMUR_OFFSET), 0x800);
  if (data_bus_size == 0x2)
  {
    reg32_write((MMDC_P1_IPS_BASE_ADDR + MPMUR_OFFSET), 0x800);
  }
  // Start the automatic write calibration process by asserting MPWRDLHWCTL0[HW_WR_DL_EN]
  reg32_write((MMDC_P0_IPS_BASE_ADDR + MPWRDLHWCTL_OFFSET), 0x00000030);
  // poll for completion
  // MMDC indicates that the write data calibration had finished by setting
  // MPWRDLHWCTL[HW_WR_DL_EN] = 0
  // Also, ensure that no error bits were set
  while (reg32_read((MMDC_P0_IPS_BASE_ADDR + MPWRDLHWCTL_OFFSET)) & 0x00000010) ;
  // check both PHYs for x64 configuration, if x32, check only PHY0
  if (data_bus_size == 0x2)
  {
    if ((reg32_read(MMDC_P0_IPS_BASE_ADDR + MPWRDLHWCTL_OFFSET) & 0x0000000f) ||
        (reg32_read(MMDC_P1_IPS_BASE_ADDR + MPWRDLHWCTL_OFFSET) & 0x0000000f))
    {
      errorcount++; // increment the errorcount variable
      g_error_write_cal = 1; // set the g_error_write_cal
    }
  } else
  {
    if (reg32_read(MMDC_P0_IPS_BASE_ADDR + MPWRDLHWCTL_OFFSET) & 0x0000000f)
    {
      errorcount++; // increment the errorcount variable
      g_error_write_cal = 1; // set the g_error_write_cal
    }
  }
  // printf("Write calibration completed\n");
  // Reset the read data FIFOs (two resets); only need to issue reset to PHY0 since in x64
  // mode, the reset will also go to PHY1
  // read data FIFOs reset1
  reg32_write((MMDC_P0_IPS_BASE_ADDR + MPDGCTRL0_OFFSET),
              reg32_read((MMDC_P0_IPS_BASE_ADDR + MPDGCTRL0_OFFSET)) | 0x80000000);
  while (reg32_read((MMDC_P0_IPS_BASE_ADDR + MPDGCTRL0_OFFSET)) & 0x80000000);
  // read data FIFOs reset2
  reg32_write((MMDC_P1_IPS_BASE_ADDR + MPDGCTRL0_OFFSET),
              reg32_read((MMDC_P1_IPS_BASE_ADDR + MPDGCTRL0_OFFSET)) | 0x80000000);
  while (reg32_read((MMDC_P1_IPS_BASE_ADDR + MPDGCTRL0_OFFSET)) & 0x80000000);
  // printf("\n");
  // enable DDR logic power down timer:
  reg32_write((MMDC_P0_IPS_BASE_ADDR + MDPDC_OFFSET),
              reg32_read((MMDC_P0_IPS_BASE_ADDR + MDPDC_OFFSET)) | 0x00005500);
  // enable Adopt power down timer:
  reg32_write((MMDC_P0_IPS_BASE_ADDR + MAPSR_OFFSET),
              reg32_read((MMDC_P0_IPS_BASE_ADDR + MAPSR_OFFSET)) & 0xfffffff7);
  //restore MDMISC value (RALAT, WALAT)
  reg32_write((MMDC_P1_IPS_BASE_ADDR + MDMISC_OFFSET), temp1);
  // device ODT back to normal:
  // reg32_write((MMDC_P0_IPS_BASE_ADDR + MPODTCTRL_OFFSET),
  // reg32_read((MMDC_P0_IPS_BASE_ADDR + MPODTCTRL_OFFSET)) & 0xfffffff7);
  //clear DQS pull ups
  reg32_write(IOMUXC_SW_PAD_CTL_PAD_DRAM_SDQS0,
              reg32_read(IOMUXC_SW_PAD_CTL_PAD_DRAM_SDQS0) & 0xffff0fff);
  reg32_write(IOMUXC_SW_PAD_CTL_PAD_DRAM_SDQS1,
              reg32_read(IOMUXC_SW_PAD_CTL_PAD_DRAM_SDQS1) & 0xffff0fff);
  reg32_write(IOMUXC_SW_PAD_CTL_PAD_DRAM_SDQS2,
              reg32_read(IOMUXC_SW_PAD_CTL_PAD_DRAM_SDQS2) & 0xffff0fff);
  reg32_write(IOMUXC_SW_PAD_CTL_PAD_DRAM_SDQS3,
              reg32_read(IOMUXC_SW_PAD_CTL_PAD_DRAM_SDQS3) & 0xffff0fff);
  reg32_write(IOMUXC_SW_PAD_CTL_PAD_DRAM_SDQS4,
              reg32_read(IOMUXC_SW_PAD_CTL_PAD_DRAM_SDQS4) & 0xffff0fff);
  reg32_write(IOMUXC_SW_PAD_CTL_PAD_DRAM_SDQS5,
              reg32_read(IOMUXC_SW_PAD_CTL_PAD_DRAM_SDQS5) & 0xffff0fff);
  reg32_write(IOMUXC_SW_PAD_CTL_PAD_DRAM_SDQS6,
              reg32_read(IOMUXC_SW_PAD_CTL_PAD_DRAM_SDQS6) & 0xffff0fff);
  reg32_write(IOMUXC_SW_PAD_CTL_PAD_DRAM_SDQS7,
              reg32_read(IOMUXC_SW_PAD_CTL_PAD_DRAM_SDQS7) & 0xffff0fff);
  // re-enable SDE (chip selects) if they were set initially
  if (cs1_enable_initial)
  {
    reg32setbit((MMDC_P0_IPS_BASE_ADDR + MDCTL_OFFSET), 30); // set SDE_1
  }
  if (cs0_enable_initial)
  {
    reg32setbit((MMDC_P0_IPS_BASE_ADDR + MDCTL_OFFSET), 31); // set SDE_0
  }
  // re-enable to auto refresh
  reg32_write((MMDC_P0_IPS_BASE_ADDR + MDREF_OFFSET), temp_ref);
  // clear the MDSCR (including the con_req bit)
  reg32_write((MMDC_P0_IPS_BASE_ADDR + MDSCR_OFFSET), 0x0); // CS0
  // poll to make sure the con_ack bit is clear
  while ((reg32_read(MMDC_P0_IPS_BASE_ADDR + MDSCR_OFFSET) & 0x00004000));
  // print out the registers that were updated as a result of the calibration process
  return errorcount;
}

void ddr3print(void)
{
  printf("MMDC registers updated from calibration \n");
  printf("\nRead DQS Gating calibration\n");
  printf("MPDGCTRL0 PHY0 (0x021b083c) = 0x%08X\n", reg32_read(0x021b083c));
  printf("MPDGCTRL1 PHY0 (0x021b0840) = 0x%08X\n", reg32_read(0x021b0840));
  printf("MPDGCTRL0 PHY1 (0x021b483c) = 0x%08X\n", reg32_read(0x021b483c));
  printf("MPDGCTRL1 PHY1 (0x021b4840) = 0x%08X\n", reg32_read(0x021b4840));
  printf("\nRead calibration\n");
  printf("MPRDDLCTL PHY0 (0x021b0848) = 0x%08X\n", reg32_read(0x021b0848));
  printf("MPRDDLCTL PHY1 (0x021b4848) = 0x%08X\n", reg32_read(0x021b4848));
  printf("\nWrite calibration\n");
  printf("MPWRDLCTL PHY0 (0x021b0850) = 0x%08X\n", reg32_read(0x021b0850));
  printf("MPWRDLCTL PHY1 (0x021b4850) = 0x%08X\n", reg32_read(0x021b4850));
  printf("\n");
  // registers below are for debugging purposes
  // these print out the upper and lower boundaries captured during read DQS gating calibration
  printf("Status registers, upper and lower bounds, for read DQS gating. \n");
  printf("MPDGHWST0 PHY0 (0x021b087c) = 0x%08X\n", reg32_read(0x021b087c));
  printf("MPDGHWST1 PHY0 (0x021b0880) = 0x%08X\n", reg32_read(0x021b0880));
  printf("MPDGHWST2 PHY0 (0x021b0884) = 0x%08X\n", reg32_read(0x021b0884));
  printf("MPDGHWST3 PHY0 (0x021b0888) = 0x%08X\n", reg32_read(0x021b0888));
  printf("MPDGHWST0 PHY1 (0x021b487c) = 0x%08X\n", reg32_read(0x021b487c));
  printf("MPDGHWST1 PHY1 (0x021b4880) = 0x%08X\n", reg32_read(0x021b4880));
  printf("MPDGHWST2 PHY1 (0x021b4884) = 0x%08X\n", reg32_read(0x021b4884));
  printf("MPDGHWST3 PHY1 (0x021b4888) = 0x%08X\n", reg32_read(0x021b4888));
}

typedef int (*DDRCalibFunction_t) (void);


int ddr_calibration(void)
{
  printf("ddr3print: %p\n", ddr3print);
  printf("ddr3cal: %p\n", ddr3cal);

  DDRCalibFunction_t Target = (void *) (IRAM_BASE_ADDR + IRAM_SIZE/2);
  memcpy(Target, ddr3cal, 4096);
  printf("ddr3cal: call at %p -> ", Target);
  printf(" %d\n", Target());
  // ddr3cal();
  ddr3print();
  return 0;
}
