/*
 * Copyright (C) 2012 Freescale Semiconductor, Inc.
 *
 * Author: Fabio Estevam <fabio.estevam@freescale.com>
 *
 * See file CREDITS for list of people who contributed to this
 * project.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 */

#define SERVERIP "248"

#include <common.h>

#include <asm/io.h>
#include <asm/arch/clock.h>
#include <asm/arch/imx-regs.h>
#include <asm/arch/iomux.h>
#include <asm/arch/mx6-pins.h>
#include <asm/errno.h>
#include <asm/gpio.h>
#include <asm/imx-common/iomux-v3.h>
#include <asm/imx-common/boot_mode.h>
#include <mmc.h>
#include <fsl_esdhc.h>
#include <miiphy.h>
#include <netdev.h>
#include <asm/imx-common/mxc_i2c.h>
#include <i2c.h>
#include <spi_flash.h>
#include <watchdog.h>
#include <fuse.h>

static int MdioWrite(uint8_t phyAddr, uint8_t regAddr, uint16_t data);
static int MdioRead(uint8_t phyAddr, uint8_t regAddr);

#include "sha1.h"

// #include "ddrcal.h"
#include "mx6smodas.h"

extern int FECPhyAddr;
static int IsSwitchDevice = 0;
static int DontDisableLANPort = 0;

#define SWITCH_CONSOLE_RESETBUTTON 0

#if 0 && defined(CONFIG_MC2)
#undef SWITCH_CONSOLE_RESETBUTTON
#define SWITCH_CONSOLE_RESETBUTTON 1
#endif

static const iomux_v3_cfg_t reset_pads[] =
{
 MX6_PAD_GPIO_2__GPIO1_IO02  | MUX_PAD_CTRL(RESET_PAD_CTRL),
};

int FoundMagic = 0;
int FoundMMC = 0;
int FoundMMCUBoot = 0;
int OverridePhySpeed = 0;

int ExecuteUCP = 0;
#pragma pack(1)

typedef unsigned char __uint8_t;
typedef unsigned short __uint16_t;
typedef unsigned int __uint32_t;
typedef __u64 __uint64_t;

typedef struct __attribute__ ((packed)) FileUpgHeader_s
{
  __uint32_t HdrMagic;
  __uint32_t HdrSize;
  __uint8_t Name[64];
  __uint8_t Data[0];
} FileUpgHeader_t;

typedef struct __attribute__ ((packed)) UpgHeader_s
{
  __uint32_t HdrMagic;
  __uint32_t HdrSize;
  __uint8_t Name[64];
  __uint8_t SoftwareVersion[64];
  __uint8_t HardwareVersion[64];
  __uint32_t UpgradeSize;
  __uint8_t Md5Sum[16];
  __uint8_t Sha1Sum[16];
  __uint32_t SignType;
  __uint8_t SignatureData[128];
  FileUpgHeader_t UpgradeFiles[0];
} UpgHeader_t;

#pragma pack()



int mmdc_do_write_level_calibration_ocram(void);
int mmdc_do_dqs_calibration_ocram(void);

typedef int (*DDRCalibFunction_t) (void);

static int PRandState = 1234;
void PRandInit(int NewState)
{
  PRandState = NewState;
}

int PRand(void)
{
  PRandState = ((PRandState * 1103515245) + 12345) & 0x7fffffff;
  return PRandState;
}

void hexdump(void *Data, int DataLen)
{
  int i = 0;
  char *DP = (char *) Data;
  while (DataLen)
  {
    printf("%02X", DP[i]);
    i++;
    if (i % 16 == 0)
    {
      printf("\r\n");
    }
    DataLen--;
  }
}


#define P0_IPS (void __iomem *)0x021b0000
#define P1_IPS (void __iomem *)0x021b4000


#define MDCTL   0x000
#define MDPDC   0x004
#define MDSCR   0x01c
#define MDMISC    0x018
#define MDREF   0x020
#define MAPSR   0x404
#define MPZQHWCTRL  0x800
#define MPWLGCR   0x808
#define MPWLDECTRL0 0x80c
#define MPWLDECTRL1 0x810
#define MPPDCMPR1 0x88c
#define MPSWDAR   0x894
#define MPRDDLCTL 0x848
#define MPMUR   0x8b8
#define MPDGCTRL0 0x83c
#define MPDGCTRL1 0x840
#define MPRDDLCTL 0x848
#define MPWRDLCTL 0x850
#define MPRDDLHWCTL 0x860
#define MPWRDLHWCTL 0x864
#define MPDGHWST0 0x87c
#define MPDGHWST1 0x880
#define MPDGHWST2 0x884
#define MPDGHWST3 0x888
int mmdc_do_write_level_calibration(void)
{
  char __attribute__((__unused__)) Test = 0;
  DDRCalibFunction_t Target = (void *) (IRAM_BASE_ADDR + IRAM_SIZE/2);
  // printf("mmdc_do_write_level_calibration_ocram: call at %p -> (Stack: %p)\r\n", Target, &Test);
  memcpy(Target, mmdc_do_write_level_calibration_ocram, 4096);
  // hexdump(Target, 64);
  int RC = Target();


  printf("MMDC_MPWLDECTRL0 after write level cal: 0x%08X\n",
      readl(P0_IPS + MPWLDECTRL0));
  //printf("MMDC_MPWLDECTRL1 after write level cal: 0x%08X\n",
//      readl(P0_IPS + MPWLDECTRL1));

  return RC;
}

int mmdc_do_dqs_calibration(void)
{
  char __attribute__((__unused__)) Test = 0;
  DDRCalibFunction_t Target = (void *) (IRAM_BASE_ADDR + IRAM_SIZE/2 + 4096);
  // printf("mmdc_do_dqs_calibration_ocram: call at %p -> (Stack: %p)\r\n", Target, &Test);
  memcpy(Target, mmdc_do_dqs_calibration_ocram, 4096);
  // hexdump(Target, 64);
  int RC = Target();

  printf("MMDC registers updated from calibration \n");
  printf("\nRead DQS Gating calibration\n");
  printf("MPDGCTRL0 PHY0 (0x021b083c) = 0x%08X\n", readl(P0_IPS + MPDGCTRL0));
  // printf("MPDGCTRL1 PHY0 (0x021b0840) = 0x%08X\n", readl(P0_IPS + MPDGCTRL1));
  // printf("MPDGCTRL0 PHY1 (0x021b483c) = 0x%08X\n", readl(P1_IPS + MPDGCTRL0));
  // printf("MPDGCTRL1 PHY1 (0x021b4840) = 0x%08X\n", readl(P1_IPS + MPDGCTRL1));
  printf("\nRead calibration\n");
  printf("MPRDDLCTL PHY0 (0x021b0848) = 0x%08X\n", readl(P0_IPS + MPRDDLCTL));
  // printf("MPRDDLCTL PHY1 (0x021b4848) = 0x%08X\n", readl(P1_IPS + MPRDDLCTL));
  printf("\nWrite calibration\n");
  printf("MPWRDLCTL PHY0 (0x021b0850) = 0x%08X\n", readl(P0_IPS + MPWRDLCTL));
  // printf("MPWRDLCTL PHY1 (0x021b4850) = 0x%08X\n", readl(P1_IPS + MPWRDLCTL));
  printf("\n");
  /*
   * registers below are for debugging purposes
   * these print out the upper and lower boundaries captured during read DQS gating calibration
   */
  printf("Status registers, upper and lower bounds, for read DQS gating. \n");
  printf("MPDGHWST0 PHY0 (0x021b087c) = 0x%08X\n", readl(P0_IPS + MPDGHWST0));
  printf("MPDGHWST1 PHY0 (0x021b0880) = 0x%08X\n", readl(P0_IPS + MPDGHWST1));
  printf("MPDGHWST2 PHY0 (0x021b0884) = 0x%08X\n", readl(P0_IPS + MPDGHWST2));
  printf("MPDGHWST3 PHY0 (0x021b0888) = 0x%08X\n", readl(P0_IPS + MPDGHWST3));
  printf("MPDGHWST0 PHY1 (0x021b487c) = 0x%08X\n", readl(P1_IPS + MPDGHWST0));
  printf("MPDGHWST1 PHY1 (0x021b4880) = 0x%08X\n", readl(P1_IPS + MPDGHWST1));
  printf("MPDGHWST2 PHY1 (0x021b4884) = 0x%08X\n", readl(P1_IPS + MPDGHWST2));
  printf("MPDGHWST3 PHY1 (0x021b4888) = 0x%08X\n", readl(P1_IPS + MPDGHWST3));


  return RC;
}


DECLARE_GLOBAL_DATA_PTR;

iomux_v3_cfg_t const pcie_pads[] =
{
 MX6_PAD_EIM_DA0__GPIO3_IO00 | MUX_PAD_CTRL(PCIE_PAD_CTRL),
 MX6_PAD_GPIO_17__GPIO7_IO12 | MUX_PAD_CTRL(PCIE_PAD_CTRL),
 MX6_PAD_CSI0_DATA_EN__GPIO5_IO20 | MUX_PAD_CTRL(PCIE_PAD_CTRL),
 MX6_PAD_KEY_COL4__GPIO4_IO14 | MUX_PAD_CTRL(PCIE_PAD_CTRL),
};

iomux_v3_cfg_t const led_pads[] =
{
 MX6_PAD_NANDF_CS1__GPIO6_IO14 | MUX_PAD_CTRL(LED_PAD_CTRL),
 MX6_PAD_NANDF_CS2__GPIO6_IO15 | MUX_PAD_CTRL(LED_PAD_CTRL),
 MX6_PAD_NANDF_CS3__GPIO6_IO16 | MUX_PAD_CTRL(LED_PAD_CTRL),
 MX6_PAD_NANDF_ALE__GPIO6_IO08  | MUX_PAD_CTRL(LED_PAD_CTRL),
 MX6_PAD_NANDF_CLE__GPIO6_IO07  | MUX_PAD_CTRL(LED_PAD_CTRL),
 MX6_PAD_NANDF_WP_B__GPIO6_IO09 | MUX_PAD_CTRL(LED_PAD_CTRL),
 MX6_PAD_NANDF_RB0__GPIO6_IO10 | MUX_PAD_CTRL(LED_PAD_CTRL),
 MX6_PAD_NANDF_D0__GPIO2_IO00 | MUX_PAD_CTRL(LED_PAD_CTRL),
 MX6_PAD_NANDF_D4__GPIO2_IO04 | MUX_PAD_CTRL(LED_PAD_CTRL),
 MX6_PAD_NANDF_D5__GPIO2_IO05 | MUX_PAD_CTRL(LED_PAD_CTRL),
 MX6_PAD_NANDF_D7__GPIO2_IO07 | MUX_PAD_CTRL(LED_PAD_CTRL),
};

void led_loop(void)
{
#if 1
  struct gpio_regs *regs = (struct gpio_regs *) GPIO6_BASE_ADDR;
  while (1)
  {
    writel(readl(&regs->gpio_dr) ^ (1 << 15), &regs->gpio_dr);
    writel(0, &regs->gpio_dr);
    writel((1 << 15), &regs->gpio_dr);
    writel(0, &regs->gpio_dr);
    writel((1 << 15), &regs->gpio_dr);
  }
#endif
}

void LEDAllOn(void)
{
  gpio_direction_output(LED1_R, 0);
  gpio_direction_output(LED1_G, 0);
  gpio_direction_output(LED1_B, 0);
  gpio_direction_output(LED2_R, 0);
  gpio_direction_output(LED2_G, 0);
  gpio_direction_output(LED2_B, 0);
  gpio_direction_output(LED3_R, 0);
  gpio_direction_output(LED3_G, 0);
  gpio_direction_output(LED3_B, 0);
  gpio_direction_output(LED4_R, 0);
  gpio_direction_output(LED4_B, 0);
}

void LEDAllOff(void)
{
  gpio_direction_output(LED1_R, 1);
  gpio_direction_output(LED1_G, 1);
  gpio_direction_output(LED1_B, 1);
  gpio_direction_output(LED2_R, 1);
  gpio_direction_output(LED2_G, 1);
  gpio_direction_output(LED2_B, 1);
  gpio_direction_output(LED3_R, 1);
  gpio_direction_output(LED3_G, 1);
  gpio_direction_output(LED3_B, 1);
  gpio_direction_output(LED4_R, 1);
  gpio_direction_output(LED4_B, 1);
}

void LEDAllBlue(void)
{
  gpio_direction_output(LED1_B, 0);
  gpio_direction_output(LED2_B, 0);
  gpio_direction_output(LED3_B, 0);
  gpio_direction_output(LED4_B, 0);
}


void LEDAllRed(void)
{
  gpio_direction_output(LED1_R, 0);
  gpio_direction_output(LED2_R, 0);
  gpio_direction_output(LED3_R, 0);
  gpio_direction_output(LED4_R, 0);
}

void LEDAllGreen(void)
{
  gpio_direction_output(LED1_G, 0);
  gpio_direction_output(LED2_G, 0);
  gpio_direction_output(LED3_G, 0);
}

void pcie_setup(void)
{
  imx_iomux_v3_setup_multiple_pads(pcie_pads, ARRAY_SIZE(pcie_pads));
  gpio_direction_output(PCIE_PWR_EN, 1);
  gpio_direction_output(PCIE_RST_B, 1);
  gpio_direction_input(PCIE_WAKE_B);
  gpio_direction_output(PCIE_DIS_B, 1);
}

void led_setup(void)
{
  imx_iomux_v3_setup_multiple_pads(led_pads, ARRAY_SIZE(led_pads));

  LEDAllOn();
}

//#define DDRTestLength 0x4000000 DDR eT²Ned: [0] 70F0DCE4 != 0890DCE4
#define DDRTestLength 0x6000000
int dram_init(void)
{
	gd->ram_size = get_ram_size((void *)PHYS_SDRAM, PHYS_SDRAM_SIZE);

#if 0
  printf("DDR3:  read-write test\r\n");

  int j;
  for (j = 0; j < 1; j++)
  {
    printf("DDR3:  pass %d\r\n", j);

    PRandInit(0x71281877);

    volatile int *pDDR = (int *) PHYS_SDRAM + 0x1000;
    int i;
    for (i = 0; (i*sizeof(int)) < DDRTestLength; i += 16)
    {
      pDDR[i] = PRand();
    }

    PRandInit(0x71281877);

    for (i = 0; (i*sizeof(int)) < DDRTestLength; i += 16)
    {
      int Expected = PRand();
      int DDRReadValue = pDDR[i];
      if (DDRReadValue != Expected)
      {
        printf("DDR Failed: [%d] %08X != %08X\r\n", i, DDRReadValue, Expected);
        udelay(1000);
        break;
      } else
      {
      }
    }
  }
#endif
	return 0;
}

/* I2C2 PMIC, iPod, Tuner, Codec, Touch, HDMI EDID, MIPI CSI2 card */
struct i2c_pads_info i2c_pad_info1 = {
  .scl = {
    .i2c_mode = MX6_PAD_KEY_COL3__I2C2_SCL | MUX_PAD_CTRL(I2C_PAD_CTRL),
    .gpio_mode = MX6_PAD_KEY_COL3__GPIO4_IO12 | MUX_PAD_CTRL(I2C_PAD_CTRL),
    .gp = IMX_GPIO_NR(4, 12)
  },
  .sda = {
    .i2c_mode = MX6_PAD_KEY_ROW3__I2C2_SDA | MUX_PAD_CTRL(I2C_PAD_CTRL),
    .gpio_mode = MX6_PAD_KEY_ROW3__GPIO4_IO13 | MUX_PAD_CTRL(I2C_PAD_CTRL),
    .gp = IMX_GPIO_NR(4, 13)
  }
};


iomux_v3_cfg_t const uart1_pads[] = {
	MX6_PAD_CSI0_DAT10__UART1_TX_DATA | MUX_PAD_CTRL(UART_PAD_CTRL),
	MX6_PAD_CSI0_DAT11__UART1_RX_DATA | MUX_PAD_CTRL(UART_PAD_CTRL),
};

iomux_v3_cfg_t const uart1_gpio_pads[] = {
  MX6_PAD_CSI0_DAT10__GPIO5_IO28 | MUX_PAD_CTRL(UART_PAD_CTRL),
  MX6_PAD_CSI0_DAT11__GPIO5_IO29 | MUX_PAD_CTRL(UART_PAD_CTRL),
};


iomux_v3_cfg_t const uart3_pads[] = {
  MX6_PAD_SD4_CLK__UART3_RX_DATA | MUX_PAD_CTRL(UART_PAD_CTRL),
  MX6_PAD_SD4_CMD__UART3_TX_DATA| MUX_PAD_CTRL(UART_PAD_CTRL),
};


#define ENET_CLK_OUTPUT (PAD_CTL_PUS_100K_DOWN |    \
  PAD_CTL_SPEED_HIGH | PAD_CTL_DSE_48ohm |     \
  PAD_CTL_HYS | PAD_CTL_SRE_FAST)

iomux_v3_cfg_t const enet_pads[] = {
  MX6_PAD_ENET_MDIO__ENET_MDIO    | MUX_PAD_CTRL(ENET_PAD_CTRL),
  MX6_PAD_ENET_MDC__ENET_MDC    | MUX_PAD_CTRL(ENET_PAD_CTRL),
  MX6_PAD_RGMII_TXC__RGMII_TXC  | MUX_PAD_CTRL(ENET_PAD_CTRL),
  MX6_PAD_RGMII_TD0__RGMII_TD0  | MUX_PAD_CTRL(ENET_PAD_CTRL),
  MX6_PAD_RGMII_TD1__RGMII_TD1  | MUX_PAD_CTRL(ENET_PAD_CTRL),
  MX6_PAD_RGMII_TD2__RGMII_TD2  | MUX_PAD_CTRL(ENET_PAD_CTRL),
  MX6_PAD_RGMII_TD3__RGMII_TD3  | MUX_PAD_CTRL(ENET_PAD_CTRL),
  MX6_PAD_RGMII_TX_CTL__RGMII_TX_CTL  | MUX_PAD_CTRL(ENET_PAD_CTRL),
  MX6_PAD_ENET_REF_CLK__ENET_TX_CLK | MUX_PAD_CTRL(ENET_PAD_CTRL),
  MX6_PAD_RGMII_RXC__RGMII_RXC  | MUX_PAD_CTRL(ENET_PAD_CTRL),
  MX6_PAD_RGMII_RD0__RGMII_RD0  | MUX_PAD_CTRL(ENET_PAD_CTRL),
  MX6_PAD_RGMII_RD1__RGMII_RD1  | MUX_PAD_CTRL(ENET_PAD_CTRL),
  MX6_PAD_RGMII_RD2__RGMII_RD2  | MUX_PAD_CTRL(ENET_PAD_CTRL),
  MX6_PAD_RGMII_RD3__RGMII_RD3  | MUX_PAD_CTRL(ENET_PAD_CTRL),
  MX6_PAD_RGMII_RX_CTL__RGMII_RX_CTL  | MUX_PAD_CTRL(ENET_PAD_CTRL),
  MX6_PAD_GPIO_16__ENET_REF_CLK | MUX_PAD_CTRL(ENET_CLK_OUTPUT),
  /* AR8031 PHY Reset */
  MX6_PAD_ENET_CRS_DV__GPIO1_IO25   | MUX_PAD_CTRL(NO_PAD_CTRL),
};

static void setup_iomux_enet(void)
{
	imx_iomux_v3_setup_multiple_pads(enet_pads, ARRAY_SIZE(enet_pads));

	enable_ent125mhz_clock();

#if 1
	/* Reset PHY */
	gpio_direction_output(IMX_GPIO_NR(1, 25), 0);
  mdelay(10);
	gpio_set_value(IMX_GPIO_NR(1, 25), 1);
  mdelay(10);
#endif

}

/*
 * 4-Bit Modus
 * Kein Card-Detect
 */
iomux_v3_cfg_t const usdhc3_pads[] =
{
	MX6_PAD_SD3_CLK__SD3_CLK      | MUX_PAD_CTRL(USDHC_PAD_CTRL),
	MX6_PAD_SD3_CMD__SD3_CMD      | MUX_PAD_CTRL(USDHC_PAD_CTRL),
	MX6_PAD_SD3_DAT0__SD3_DATA0   | MUX_PAD_CTRL(USDHC_PAD_CTRL),
	MX6_PAD_SD3_DAT1__SD3_DATA1   | MUX_PAD_CTRL(USDHC_PAD_CTRL),
	MX6_PAD_SD3_DAT2__SD3_DATA2   | MUX_PAD_CTRL(USDHC_PAD_CTRL),
	MX6_PAD_SD3_DAT3__SD3_DATA3   | MUX_PAD_CTRL(USDHC_PAD_CTRL),
};

static void setup_iomux_uart(void)
{
#if CONFIG_CONSOLE_DEV_INDEX == 2
  if (!gpio_get_value(GPIO_RESET) && SWITCH_CONSOLE_RESETBUTTON)
  {
    imx_iomux_v3_setup_multiple_pads(uart1_pads, ARRAY_SIZE(uart1_pads));
  } else
  {
    imx_iomux_v3_setup_multiple_pads(uart1_gpio_pads, ARRAY_SIZE(uart1_gpio_pads));
  }
  imx_iomux_v3_setup_multiple_pads(uart3_pads, ARRAY_SIZE(uart3_pads));
#endif
#if CONFIG_CONSOLE_DEV_INDEX == 0
  imx_iomux_v3_setup_multiple_pads(uart1_pads, ARRAY_SIZE(uart1_pads));
#endif
}

struct fsl_esdhc_cfg usdhc_cfg[1] =
{
	{USDHC3_BASE_ADDR},
};

/*
 * Erkennung ob eine Karte steckt -> Gibt es hier nicht, also immer
 * OK melden und dann sehen ob eine Kommunikation mit der Karte möglich ist.
 */
int board_mmc_getcd(struct mmc *mmc)
{
	struct fsl_esdhc_cfg *cfg = (struct fsl_esdhc_cfg *)mmc->priv;
	int ret = 0;

	switch (cfg->esdhc_base)
	{
	  case USDHC3_BASE_ADDR:
	  {
	    ret = 1;
	    break;
	  }
	}

	return ret;
}

int board_mmc_init(bd_t *bis)
{
	s32 status = 0;
	int i;

	/*
	 * According to the board_mmc_init() the following map is done:
	 * mmc0                    SD3
	 */
	for (i = 0; i < CONFIG_SYS_FSL_USDHC_NUM; i++)
	{
	  switch (i)
	  {
	    case 0:
	    {
	      imx_iomux_v3_setup_multiple_pads(usdhc3_pads, ARRAY_SIZE(usdhc3_pads));
	      usdhc_cfg[0].sdhc_clk = mxc_get_clock(MXC_ESDHC3_CLK);
	      usdhc_cfg[0].sdhc_clk = MIN(20000000, usdhc_cfg[0].sdhc_clk);
	      break;
	    }
	    default:
	      printf("Warning: you configured more USDHC controllers"
	          "(%d) then supported by the board (%d)\n",
	          i + 1, CONFIG_SYS_FSL_USDHC_NUM);
	      return status;
	  }
	  status |= fsl_esdhc_initialize(bis, &usdhc_cfg[i]);
	}
	return status;
}

int ConfigAtherosPhy(struct phy_device *phydev)
{
	unsigned short val;

	/* To enable AR8031 ouput a 125MHz clk from CLK_25M */
	phy_write(phydev, MDIO_DEVAD_NONE, 0xd, 0x7);
	phy_write(phydev, MDIO_DEVAD_NONE, 0xe, 0x8016);
	phy_write(phydev, MDIO_DEVAD_NONE, 0xd, 0x4007);

	val = phy_read(phydev, MDIO_DEVAD_NONE, 0xe);
	val &= 0xffe3;
	val |= 0x18;
	phy_write(phydev, MDIO_DEVAD_NONE, 0xe, val);

	/* introduce tx clock delay */
	phy_write(phydev, MDIO_DEVAD_NONE, 0x1d, 0x5);
	val = phy_read(phydev, MDIO_DEVAD_NONE, 0x1e);
	val |= 0x0100;
	phy_write(phydev, MDIO_DEVAD_NONE, 0x1e, val);

  LEDAllGreen();

	return 0;
}

__uint32_t QCSwitchRead32(int Address)
{
  __uint16_t ValLo;
  __uint16_t ValHi;
  Address = Address/2;
  miiphy_read("FEC", 0x10, Address + 0, &ValLo);
  miiphy_read("FEC", 0x10, Address + 1, &ValHi);
  return ((__uint32_t) ValHi << 16) | ((__uint32_t) ValLo);
}

void QCSwitchWrite32(int Address, __uint32_t Value)
{
  __uint16_t ValLo;
  __uint16_t ValHi;
  Address = Address/2;
  miiphy_write("FEC", 0x10, Address + 0, (Value >>  0) & 0xFFFF);
  miiphy_write("FEC", 0x10, Address + 1, (Value >> 16) & 0xFFFF);
}

void QCSwitchSetClearBits32(int Address, __uint32_t SetValue, __uint32_t ClearValue)
{
  __uint32_t OldValue = QCSwitchRead32(Address);
  OldValue |= SetValue;
  OldValue &= ~ClearValue;
  QCSwitchWrite32(Address, OldValue);
}

#define MAC0_RGMII_EN (1 << 26)
#define MAC0_RGMII_TXCLK_DELAY_EN (1 << 25)
#define MAC0_RGMII_TXCLK_DEL(N)   ((N << 22) | MAC0_RGMII_TXCLK_DELAY_EN)
#define MAC0_RGMII_RXCLK_DELAY_EN (1 << 24)
#define MAC0_RGMII_RXCLK_DEL(N)   ((N << 20) | MAC0_RGMII_RXCLK_DELAY_EN)
int ConfigQCSwitch(struct phy_device *phydev)
{
  int val;

  const char *devname;
  unsigned char phyaddr = 3;
  unsigned char reg = 0;
  unsigned short  data;

  // printf("\nDevice = %08X\n", QCSwitchRead32(0));
  QCSwitchWrite32(0x4, MAC0_RGMII_EN | MAC0_RGMII_TXCLK_DEL(1)); // | MAC0_RGMII_RXCLK_DEL(3)); // | (1 << 12) | (1 << 8) | (1 << 4) | 1);

#if 0
  printf("\nMode = %08X\n", QCSwitchRead32(0x04));

  printf("\nLEDCtrl = %08X\n", QCSwitchRead32(0x50));
  printf("\nLEDCtrl = %08X\n", QCSwitchRead32(0x54));
  printf("\nLEDCtrl = %08X\n", QCSwitchRead32(0x58));
  printf("\nLEDCtrl = %08X\n", QCSwitchRead32(0x5C));
#endif
  // QCSwitchWrite32(0x5C, 0x0);

  // PORT0_STATUS
  QCSwitchSetClearBits32(0x7C, (1 << 1) | (1 << 2) | (1 << 4) | (1 << 5), 1 << 0);
#if 0
  printf("\nPORT0_STATUS = %08X\n", QCSwitchRead32(0x7C));
  printf("PORT2_STATUS = %08X\n", QCSwitchRead32(0x84));
  printf("PORT3_STATUS = %08X\n", QCSwitchRead32(0x88));
#endif

  // RX Clock Delay!
#if 0
  phy_write(phydev, MDIO_DEVAD_NONE, 0x1d, 0x0);
  val = phy_read(phydev, MDIO_DEVAD_NONE, 0x1e);
  val |= 0x8000;
  phy_write(phydev, MDIO_DEVAD_NONE, 0x1e, val);
#endif

  QCSwitchSetClearBits32(0, 1 << 31, 0);

  __uint16_t MII_BCMR;
  miiphy_read("FEC", 0, 0, &MII_BCMR);
  // printf("MII_BCMR = %08X\n", MII_BCMR);
  MII_BCMR &= ~(1 << 12);
  MII_BCMR &= ~(1 << 13);
  MII_BCMR &= ~(1 << 9);
  MII_BCMR = (1 << 8) | (1 << 6);
  miiphy_write("FEC", 0, 0, MII_BCMR);
  // printf("MII_BCMR = %08X\n", MII_BCMR);
  MII_BCMR |= (1 << 9);
  miiphy_write("FEC", 0, 0, MII_BCMR);
  // printf("MII_BCMR = %08X\n", MII_BCMR);
  miiphy_read("FEC", 0, 0, &MII_BCMR);
  // printf("MII_BCMR = %08X\n", MII_BCMR);


  return 0;
}

int board_phy_config(struct phy_device *phydev)
{
  int PhyID;
  int PhyID2;
  PhyID = phy_read(phydev, MDIO_DEVAD_NONE, 2);
  PhyID2 = phy_read(phydev, MDIO_DEVAD_NONE, 3);
  printf("\nFound phy device: %04X %04X\n", PhyID, PhyID2);
  if (PhyID2 == 0xFFFF || PhyID == 0xFFFF) return -1;

  if (IsSwitchDevice == 1)
  {
    ConfigQCSwitch(phydev);
  } else
  {
    ConfigAtherosPhy(phydev);
  }
  if (phydev->drv->config)
    phydev->drv->config(phydev);

  if (IsSwitchDevice == 1 && !DontDisableLANPort)
  {
    u16 MII_BCMR;
    int i;
    for (i = 1; i <= 5; i++)
    {
      MdioWrite(i, 0, MdioRead(i, 0) | BMCR_PDOWN);
    }
  }
	return 0;
}


/* MII-related definitios */
#define FEC_MII_DATA_ST   0x40000000  /* Start of frame delimiter */
#define FEC_MII_DATA_OP_RD  0x20000000  /* Perform a read operation */
#define FEC_MII_DATA_OP_WR  0x10000000  /* Perform a write operation */
#define FEC_MII_DATA_PA_MSK 0x0f800000  /* PHY Address field mask */
#define FEC_MII_DATA_RA_MSK 0x007c0000  /* PHY Register field mask */
#define FEC_MII_DATA_TA   0x00020000  /* Turnaround */
#define FEC_MII_DATA_DATAMSK  0x0000ffff  /* PHY data field */

#define FEC_MII_DATA_RA_SHIFT 18  /* MII Register address bits */
#define FEC_MII_DATA_PA_SHIFT 23  /* MII PHY address bits */
#define FEC_IEVENT_MII      0x00800000

/*
 * MII-interface related functions
 */
static int MdioRead(uint8_t phyAddr, uint8_t regAddr)
{
  uint32_t reg;   /* convenient holder for the PHY register */
  uint32_t phy;   /* convenient holder for the PHY */
  uint32_t start;
  int val;

  /*
   * reading from any PHY's register is done by properly
   * programming the FEC's MII data register.
   */
  writel(FEC_IEVENT_MII, 0x02188004);
  reg = regAddr << FEC_MII_DATA_RA_SHIFT;
  phy = phyAddr << FEC_MII_DATA_PA_SHIFT;

  writel(FEC_MII_DATA_ST | FEC_MII_DATA_OP_RD | FEC_MII_DATA_TA |
      phy | reg, 0x02188040);

  /*
   * wait for the related interrupt
   */
  start = get_timer(0);
  while (!(readl(0x02188004) & FEC_IEVENT_MII)) {
    if (get_timer(start) > (CONFIG_SYS_HZ / 1000)) {
      printf("Read MDIO failed...\n");
      return -1;
    }
  }

  /*
   * clear mii interrupt bit
   */
  writel(FEC_IEVENT_MII, 0x02188004);

  /*
   * it's now safe to read the PHY's register
   */
  val = (unsigned short)readl(0x02188040);
  return val;
}

static int MdioWrite(uint8_t phyAddr, uint8_t regAddr, uint16_t data)
{
  uint32_t reg;   /* convenient holder for the PHY register */
  uint32_t phy;   /* convenient holder for the PHY */
  uint32_t start;

  reg = regAddr << FEC_MII_DATA_RA_SHIFT;
  phy = phyAddr << FEC_MII_DATA_PA_SHIFT;

  writel(FEC_MII_DATA_ST | FEC_MII_DATA_OP_WR |
    FEC_MII_DATA_TA | phy | reg | data, 0x02188040);

  /*
   * wait for the MII interrupt
   */
  start = get_timer(0);
  while (!(readl(0x02188004) & FEC_IEVENT_MII)) {
    if (get_timer(start) > (CONFIG_SYS_HZ / 1000)) {
      printf("Write MDIO failed...\n");
      return -1;
    }
  }

  /*
   * clear MII interrupt bit
   */
  writel(FEC_IEVENT_MII, 0x02188004);

  return 0;
}


int board_eth_init(bd_t *bis)
{
  int Phy1ID;
  int Phy1ID2;
  int Phy4ID;
  int Phy4ID2;

  int ret;
  int val;

	setup_iomux_enet();

	writel(00000030, 0x02188044);

	Phy1ID = MdioRead(1, 2);
	Phy1ID2 = MdioRead(1, 3);

	Phy4ID = MdioRead(4, 2);
	Phy4ID2 = MdioRead(4, 3);

	if (Phy4ID == 0x004D && Phy4ID2 == 0xD072)        // Single PHY?
	{
	  FECPhyAddr = 4;
	} else if (Phy4ID == 0x004D && Phy4ID2 == 0xD036) // Switch?
	{
	  IsSwitchDevice = 1;
#ifdef CONFIG_DFU5
    FECPhyAddr = 4;
#else
	  FECPhyAddr = 1;
#endif
	  OverridePhySpeed = 1000;
	} else
	{
    printf("Unknown PHY!\n");
	}

  printf("FEC MXC: Phy ID %d\n", FECPhyAddr);

	ret = cpu_eth_init(bis);
	if (ret)
	{
#if 0
	  FECPhyAddr = 4;
	  ret = cpu_eth_init(bis);
#endif
		printf("FEC MXC: %s:failed\n", __func__);
	}

	int IOMUXC_GPR1 = readl(0x20E0004);
	writel(IOMUXC_GPR1 | (1 << 21), 0x20E0004);

#if 0
  printf("\n0x20C80E0 = %08X (CCM_ANALOG_PLL_ENETn)\n", readl(0x20C80E0));
  printf("0x20E0214 = %08X (IOMUXC_SW_MUX_CTL_PAD_GPIO16)\n", readl(0x20E0214));
  printf("0x20E05E4 = %08X (IOMUXC_SW_PAD_CTL_PAD_GPIO16)\n", readl(0x20E05E4));
  printf("0x20E0004 = %08X (IOMUXC_GPR1) %d\n", readl(0x20E0004), !!(readl(0x20E0004) & (1 << 21)));
  printf("0x20E05C0 = %08X (IOMUXC_SW_PAD_CTL_PAD_ENET_REF_CLK)\n", readl(0x20E05C0));
  printf("0x20E080C = %08X (IOMUXC_ENET_REF_CLK_SELECT_INPUT)\n", readl(0x20E080C));
#endif

	return 0;
}

#define USBOTG_IN_PAD_CTRL  (PAD_CTL_HYS | PAD_CTL_PUS_100K_DOWN | PAD_CTL_DSE_240ohm | PAD_CTL_SRE_SLOW)
#define USBOTG_OUT_PAD_CTRL (                                      PAD_CTL_DSE_240ohm | PAD_CTL_SRE_SLOW)

static void setup_iomux_usbotg(void)
{
  static const iomux_v3_cfg_t usbotg_pads[] =
  {
   MX6_PAD_EIM_D22__USB_OTG_PWR  | MUX_PAD_CTRL(USBOTG_OUT_PAD_CTRL),
   MX6_PAD_EIM_D21__USB_OTG_OC  | MUX_PAD_CTRL(USBOTG_IN_PAD_CTRL),
   MX6_PAD_ENET_RX_ER__USB_OTG_ID  | MUX_PAD_CTRL(USBOTG_IN_PAD_CTRL),
  };

  /* Set up pins for USBOTG. */
  imx_iomux_v3_setup_multiple_pads(usbotg_pads, ARRAY_SIZE(usbotg_pads));
}

int RelStartState = 0;
int board_early_init_f(void)
{
  imx_iomux_v3_setup_multiple_pads(reset_pads, ARRAY_SIZE(reset_pads));
  gpio_direction_input(GPIO_RESET);

#ifdef CONFIG_HW_WATCHDOG
  hw_watchdog_init();
#endif

  pcie_setup();
  led_setup();

  setup_iomux_usbotg();
	setup_iomux_uart();
  printf("Debug...\n");
	return 0;
}


iomux_v3_cfg_t const ecspi3_pads[] =
{
 MX6_PAD_DISP0_DAT3__GPIO4_IO24  | MUX_PAD_CTRL(SPI_PAD_CTRL), /* SS0 */
 MX6_PAD_DISP0_DAT2__ECSPI3_MISO | MUX_PAD_CTRL(SPI_PAD_CTRL),
 MX6_PAD_DISP0_DAT1__ECSPI3_MOSI | MUX_PAD_CTRL(SPI_PAD_CTRL),
 MX6_PAD_DISP0_DAT0__ECSPI3_SCLK | MUX_PAD_CTRL(SPI_PAD_CTRL),
};

void setup_spi(void)
{
  puts("SPI:   Init pad configuration (ECSPI3, CS0)\n");
  imx_iomux_v3_setup_multiple_pads(ecspi3_pads, ARRAY_SIZE(ecspi3_pads));
}

int board_init(void)
{
	/* address of boot parameters */
	gd->bd->bi_boot_params = PHYS_SDRAM + 0x100;

  setup_spi();

  return 0;
}

#define PFUZE100_SW1ABVOL 0x20
#define PFUZE100_SW1CVOL  0x2e
#define PFUZE100_SW2VOL   0x35
#define PFUZE100_SW3AVOL  0x3c
#define PFUZE100_SW3BVOL  0x43
#define PFUZE100_SW4VOL   0x4a

struct PFUZE100VoltageRange_s
{
  int Min;
  int Max;
  int Inc;
};

struct PFUZE100Voltages_s
{
  char *Name;
  int Reg;
  int CorValue;
  struct PFUZE100VoltageRange_s LowRange;
  struct PFUZE100VoltageRange_s HighRange;
};

#define PFUZE100_SW_REG(_Name, _RegCorrValue, _RegVolt, _MinLow, _MaxLow, _IncrementLow, _MinHigh, _MaxHigh, _IncrementHigh)            \
  {                                                                                        \
    .Name = #_Name, .Reg = _RegVolt, .CorValue = _RegCorrValue,                            \
    .LowRange.Min = _MinLow, .LowRange.Max = _MaxLow, .LowRange.Inc = _IncrementLow,       \
    .HighRange.Min = _MinHigh, .HighRange.Max = _MaxHigh, .HighRange.Inc = _IncrementHigh, \
  }

static struct PFUZE100Voltages_s PFUZE100Voltages[] =
{
  PFUZE100_SW_REG(SW1AB, 0x00, PFUZE100_SW1ABVOL,  300000, 1075000, 25000, 1100000, 1875000, 25000),
  PFUZE100_SW_REG(SW1C,  0x00, PFUZE100_SW1CVOL,   300000, 1075000, 25000, 1100000, 1875000, 25000),
  PFUZE100_SW_REG(SW2,   0x00, PFUZE100_SW2VOL,    400000, 1975000, 25000, 800000,  3300000, 50000),
  PFUZE100_SW_REG(SW3A,  0x00, PFUZE100_SW3AVOL,   400000, 1975000, 25000, 800000,  3300000, 50000),
  PFUZE100_SW_REG(SW3B,  0x00, PFUZE100_SW3BVOL,   400000, 1975000, 25000, 800000,  3300000, 50000),
  PFUZE100_SW_REG(SW4,   0x20, PFUZE100_SW4VOL,    400000, 1975000, 25000, 800000,  3300000, 50000),
};


int pfuze100_init(void)
{
  setup_i2c(1, CONFIG_SYS_I2C_SPEED, 0x7F, &i2c_pad_info1);

  unsigned char PfuzeRegVal;

  if (i2c_read(CONFIG_SYS_I2C_PMIC_PFUZE100_ADDR, 0xB8, 1, (void *) &PfuzeRegVal, sizeof(PfuzeRegVal)) == 0)
  {
    if (PfuzeRegVal == 0x20)
    {
      return 0;
    }
  }

  // Setzen der alternativspannungen wenn der Defaultwert nicht stimmt dauert ca 350ms
  int i;
  for (i = 0; i < sizeof(PFUZE100Voltages)/sizeof(PFUZE100Voltages[0]); i++)
  {
    struct PFUZE100Voltages_s VoltageDef = PFUZE100Voltages[i];
    if (i2c_read(CONFIG_SYS_I2C_PMIC_PFUZE100_ADDR, VoltageDef.Reg, 1, (void *) &PfuzeRegVal, sizeof(PfuzeRegVal)) == 0)
    {
      struct PFUZE100VoltageRange_s Range = VoltageDef.LowRange;
      if (PfuzeRegVal & 0x40) { Range = VoltageDef.HighRange; }
      printf("PMIC: Default %s voltage %dmV (%02X)", VoltageDef.Name, Range.Min + Range.Inc*(PfuzeRegVal & 0x3F), PfuzeRegVal);
      if (VoltageDef.CorValue && PfuzeRegVal != VoltageDef.CorValue)
      {
        PfuzeRegVal = VoltageDef.CorValue;
        if (i2c_write(CONFIG_SYS_I2C_PMIC_PFUZE100_ADDR, VoltageDef.Reg, 1, (void *) &PfuzeRegVal, sizeof(PfuzeRegVal)) == 0)
        {
        }
        if (PfuzeRegVal & 0x40) { Range = VoltageDef.HighRange; }
        printf(" changed value to %s: %dmV", VoltageDef.Name, Range.Min + Range.Inc*(PfuzeRegVal & 0x3F));
      }
      printf("\n");
    }
  }

  if (i2c_read(CONFIG_SYS_I2C_PMIC_PFUZE100_ADDR, 0xB8, 1, (void *) &PfuzeRegVal, sizeof(PfuzeRegVal)) == 0)
  {
    printf("0xB8: 0x%02X\r\n", PfuzeRegVal);
    if (PfuzeRegVal == 0x20)
    {
      return 0;
    }
  }
  if (i2c_read(CONFIG_SYS_I2C_PMIC_PFUZE100_ADDR, 0xE4, 1, (void *) &PfuzeRegVal, sizeof(PfuzeRegVal)) == 0)
  {
    printf("0xE4: 0x%02X\r\n", PfuzeRegVal);
  }
  PfuzeRegVal = 0x01;
  if (i2c_write(CONFIG_SYS_I2C_PMIC_PFUZE100_ADDR, 0x7F, 1, (void *) &PfuzeRegVal, sizeof(PfuzeRegVal)) == 0)
  {
    printf("changed 0x7F -> 0x01\r\n");
  }
  PfuzeRegVal = 0x20;
  if (i2c_write(CONFIG_SYS_I2C_PMIC_PFUZE100_ADDR, 0xB8, 1, (void *) &PfuzeRegVal, sizeof(PfuzeRegVal)) == 0)
  {
    printf("changed 0xB8 -> 0x20\r\n");
  }
  PfuzeRegVal = 0x80;
  if (i2c_write(CONFIG_SYS_I2C_PMIC_PFUZE100_ADDR, 0xE4, 1, (void *) &PfuzeRegVal, sizeof(PfuzeRegVal)) == 0)
  {
    printf("changed 0xE4 -> 0x80\r\n");
  }
  gpio_direction_output(GPIO_RESET, 0);

  return 0;
}


static int init_mmc_for_env(struct mmc *mmc)
{
  if (!mmc) {
    puts("No MMC card found\n");
    return -1;
  }

  if (mmc_init(mmc)) {
    puts("MMC init failed\n");
    return -1;
  }

  return 0;
}


void UCPExecuteFirmwareUpgrade(void *Addr, size_t Len)
{
  printf("UCPExecuteFirmwareUpgrade -> %p %d\n", Addr, Len);
  UpgHeader_t *UpgData = (UpgHeader_t *) Addr;
  printf("Upgrade Magic: %08X\n", UpgData->HdrMagic);
  printf("Upgrade Size: %08X\n", UpgData->UpgradeSize);
  printf("Upgrade Name: %s\n", UpgData->Name);

  __uint8_t Sha1Sum[SHA1_SUM_LEN];

  const unsigned char *UpgradeData = Addr + sizeof(UpgHeader_t);
  int UpgradeDataLen = Len - sizeof(UpgHeader_t);
  sha1_csum(UpgradeData, UpgradeDataLen, Sha1Sum);

  if (memcmp(Sha1Sum, UpgData->Sha1Sum, SHA1_SUM_LEN) == 0)
  {
    printf("SHA1 OK!\n");
  } else
  {
    printf("SHA1 Failed!\n");
    int i;
    printf("SHA1 Calc: ");
    for (i = 0; i < SHA1_SUM_LEN; i++)
    {
      printf("%02X", Sha1Sum[i]);
    }
    printf("\n");

    printf("SHA1 Header: ");
    for (i = 0; i < SHA1_SUM_LEN; i++)
    {
      printf("%02X", UpgData->Sha1Sum[i]);
    }
    printf("\n");
  }

  int Offset = 1024;
  if (strcmp("UBootIMX6", (char *) UpgData->Name) == 0)
  {
    printf("UBoot upgrade\n");
    Offset = 1024;
  } else
  {
    printf("Unknown Upgrade Type...\n");
    return;
  }

#ifdef CONFIG_ENV_IS_IN_SPI_FLASH

#endif

#ifdef CONFIG_ENV_IS_IN_MMC
  struct mmc *MMCDev = find_mmc_device(CONFIG_SYS_MMC_ENV_DEV);
  if (init_mmc_for_env(MMCDev)) return;

  const unsigned char *Buffer = UpgradeData;
  int Size = UpgradeDataLen;
  uint BlkStart, BlkCount, n;

  BlkStart = ALIGN(Offset, MMCDev->write_bl_len) / MMCDev->write_bl_len;
  BlkCount = ALIGN(Size, MMCDev->write_bl_len) / MMCDev->write_bl_len;
  printf("Write to MMC offset %08X BlockCount %08X!\n", BlkStart, BlkCount);
  n = MMCDev->block_dev.block_write(CONFIG_SYS_MMC_ENV_DEV, BlkStart, BlkCount, (u_char *) Buffer);

  if (n != BlkCount)
  {
    printf("Write to MMC Failed!\n");
  } else
  {
    printf("Write to MMC OK!\n");
    do_reset(NULL, 0, 0, NULL);
  }
#endif
}

struct watchdog_regs {
  u16 wcr;  /* Control */
  u16 wsr;  /* Service */
  u16 wrsr; /* Reset Status */
};

int misc_init_r(void)
{
 // pfuze100_init();


  struct watchdog_regs *wdog = (struct watchdog_regs *) WDOG1_BASE_ADDR;
  if (wdog->wrsr & (1 << 1))
  {
    gpio_direction_output(GPIO_RESET, 0);
  }


#if 1
  static const iomux_v3_cfg_t rel_pads[] =
  {
   MX6_PAD_GPIO_4__GPIO1_IO04  | MUX_PAD_CTRL(REL_PAD_CTRL),
  };
  gpio_direction_input(GPIO_REL);
  RelStartState = gpio_get_value(GPIO_REL);
  imx_iomux_v3_setup_multiple_pads(rel_pads, ARRAY_SIZE(rel_pads));
  gpio_direction_output(GPIO_REL, RelStartState);
  gpio_set_value(GPIO_REL, RelStartState);
#endif

#if BOOT_SETFUSES
  u32 bank = 0;
  u32 word = 0;
  u32 val;
  int i;
  for (i = 0; i < 8; i++)
  {
    word = i;
    int ret = fuse_sense(bank, word, &val);
    printf("fuse %02X[%02X]: %08X\r\n", bank, word, val);
  }

  u32 UniqueID1;
  u32 UniqueID2;
  u32 ret = fuse_sense(0, 1, &UniqueID1);
  ret = fuse_sense(0, 2, &UniqueID2);
  printf("UniqueID: %08X%08X\r\n", UniqueID1, UniqueID2);

  u32 MAC1;
  u32 MAC2;
  ret = fuse_sense(4, 2, &MAC1);
  ret = fuse_sense(4, 3, &MAC2);
  printf("MAC: %04X%08X\r\n", MAC2, MAC1);

  // 90:5f:8d:
  if (!MAC1 && !MAC2)
  {
    fuse_prog(4, 2, 0x8D000000);
    fuse_prog(4, 3, 0x0000905F);
  }

  u32 BootREG1;
  u32 BootREG2;
  ret = fuse_sense(0, 5, &BootREG1);
  ret = fuse_sense(0, 6, &BootREG2);
  printf("BOOT_CFG: %08X\r\n", BootREG1);
  printf("BT_FUSE_SEL[4]: %08X\r\n", BootREG2);
  if (!BootREG1)
  {
    // CFG1[7:4] = 0011 XXXX : *Boot from Serial ROM*
    // CFG4[5:0] = XX 00 1 010 : ECSPIx_*SS0* ECSPI-*3* *24-Bit* Adressing
#ifdef BOOT_DEVICE_SPI
    fuse_prog(0, 5, 0x0A000030);
#endif

#if 0
    EMMC Boot eFUSE Configuration
                | *Boot vom USDHC Interface*
                   | 0 = SD-Card, 1 = *MMC*
                     | 0 = *Normal Boot*, 1 = Fast Boot       <------------ ?
                       | 0X = *High Speed Mode*, 1x = Normal Mode
                          | 0 = *eMMC reset disabled*, 1 = ...
    CFG1[7:0] = 01 1 0 0X 0 0

                | 001 = *Bus Width 4-Bit*, ....
                    | 10 = *USDHC-3*
                       | 0 = Boot ROM default
                         | 0 = Boot Acknowledge Enabled
                           | 0 = Use default values (PAD-settings...)
    CFG2[7:0] = 001 10 0 0 0
#endif
#ifdef BOOT_DEVICE_EMMC
    fuse_prog(0, 5, 0x00003060);
#endif

#if 0

    SD-Card Boot eFUSE Configuration
                | *Boot vom USDHC Interface*
                   | 0 = *SD-Card*, 1 = MMC
                     | 0 = *Normal Boot*, 1 = Fast Boot       <------------ ?
                       | *0x = High/Normal*, 10 = SDR50, 11 = SDR104
                          | 0 = *SD no power cycle*, 1 = Power cycle via SD_RST pad
    CFG1[7:0] = 01 0 0 0X 0 0

                | XX0 = *Bus Width 1-Bit*, ....
                    | 10 = *USDHC-3*
                       | 0 = Boot ROM default
                         | 0 = use default SD pad settings during power cycle
                           | 0 = Use default values (PAD-settings...)
    CFG2[7:0] = XX0 10 0 0 0
#endif

#ifdef BOOT_DEVICE_SDCARD
    fuse_prog(0, 5, 0x00001040);
#endif

  }
  if (!BootREG2) // BT_FUSE_SEL[4] -> Use Fuses to boot!
  {
#if defined(BOOT_DEVICE_SPI) || defined(BOOT_DEVICE_EMMC) || defined(BOOT_DEVICE_SDCARD)
   fuse_prog(0, 6, 0x00000010);
#endif
  }
#endif

#if 0
  U-Boot > fuse prog 0 5 0x18000030     // BOOT_CFG[4..1]
  U-Boot > fuse prog 0 6 0x00000010     // BT_FUSE_SEL = 1
#endif

  gpio_direction_input(GPIO_RESET);

  LEDAllOff();

#if 0
  // System-Data auslesen. Magic steht ganz am Anfang:
  Type 5f fa
  Length 08 00
  Value 41 54 41 44  5f 53 59 53 (ATAD_SYS)

  root@OpenWrt:/# hexdump /dev/mtd0 -s 786432 -n 20 -C
  000c0000  5f fa 08 00 41 54 41 44  5f 53 59 53 00 00 f0 04  |_...ATAD_SYS....|
  000c0010  00 83 6d 0d                                       |..m.|
  000c0014
#endif

  // Reset Env: env default -f -a

  // Cmdline (bootargs) ist unter linux unter /proc/cmdline auslesbar
  char *Args = getenv("bootargs");
#if CONFIG_CONSOLE_DEV_INDEX == 2
  if (Args == 0) Args = "console=ttymxc2,115200 ath9k.ps_enable=1 pci=nomsi ";
#endif
#if CONFIG_CONSOLE_DEV_INDEX == 0
  if (Args == 0) Args = "console=ttymxc0,115200 ath9k.ps_enable=1 pci=nomsi ";
#endif
  char TmpBootArgs[1024];
  if (Args)
  {
    if (strstr(Args, " SD="))
    {
      sprintf(TmpBootArgs, "%s", Args);
    } else
    {
      sprintf(TmpBootArgs, "%s SD=mtd0,000C0000", Args);
    }
    char *TTYPos = strstr(TmpBootArgs, "ttymxc");
    if (TTYPos)
    {
      if (!gpio_get_value(GPIO_RESET) && SWITCH_CONSOLE_RESETBUTTON)
      {
        TTYPos[6] = '0'; // Nur wenn reset gerückt ist debug auf die normale Schnittstelle
      } else
      {
#if CONFIG_CONSOLE_DEV_INDEX == 2
        TTYPos[6] = '2'; // Andernfalls debug nur auf die interne!
#endif
//        memcpy(TTYPos, "quiet         ", 14);
      }
    }

    u32 UniqueID1;
    u32 UniqueID2;
    u32 __attribute__((__unused__)) ret = fuse_sense(0, 1, &UniqueID1);
    ret = fuse_sense(0, 2, &UniqueID2);
    if (UniqueID1 || UniqueID2)
    {
      sprintf(&TmpBootArgs[strlen(TmpBootArgs)], " uuid=%08X%08X ", UniqueID1, UniqueID2);
    }

    sprintf(&TmpBootArgs[strlen(TmpBootArgs)], "rel=%d ", RelStartState);
    sprintf(&TmpBootArgs[strlen(TmpBootArgs)], "v=" __DATE__ __TIME__);

    struct watchdog_regs *wdog = (struct watchdog_regs *) WDOG1_BASE_ADDR;
    sprintf(&TmpBootArgs[strlen(TmpBootArgs)], " wrsr=%04X ", wdog->wrsr);

    setenv("bootargs", TmpBootArgs);
    printf("Boot:  %s\r\n", TmpBootArgs);
  }

  setenv("bootdelay", "10");

  InitSystemData();
#if BOOT_SETFUSES

  printf("!FoundMagic %d && !FoundMMCUBoot %d && FoundMMC %d\n", !FoundMagic, !FoundMMCUBoot, FoundMMC);
  // Test: mmc erase 0x00000000 0x00200000
  if (!FoundMagic && !FoundMMCUBoot && FoundMMC)
  {
    setenv("bootdelay", "0");
    setenv("serverip", DEFAULT_SUBNET SERVERIP);
    setenv("ipaddr", DEFAULT_SUBNET "111");
    setenv("LoadII", "mmc read 0x13000000 0x00005000 0x00005000; "
           "tftp 0x12000000 uImage && mmc erase 0x00000800 0x00004000 && mmc write 0x12000000 0x00000800 0x00004000 && "
           "tftp 0x13000000 CpioRoot && mmc erase 0x00005000 0x00005000 && mmc write 0x13000000 0x00005000 0x00005000 && "
           "tftp 0x12000000 u-boot.imx &&    mmc erase 0x00000000 0x00000400 && mmc write 0x12000000 0x00000002 0x00000300 && "
           "tftp 0x14000000 DTB &&      mmc erase 0x0000A000 0x00000400 && mmc write 0x14000000 0x0000A000 0x00000400 && reset");
    setenv("bootcmd", "run LoadII");
    DontDisableLANPort = 1;
    return 0;
  }
#endif

  if (!gpio_get_value(GPIO_RESET))
  {
    int i;
    DontDisableLANPort = 1;
    printf("Pressed reset and holding -> Failsafe booting... (Wait for choice)!\r\n");
    int Option = 3;
    int LoopCount = 0;
    while (!gpio_get_value(GPIO_RESET))
    {
      Option = (Option + 1) % 4;
      LEDAllOff();
#if !(defined(CONFIG_BWLC)) && !(defined(CONFIG_MC6))
      switch (Option)
      {
        case 0:
          LoopCount++;
          LEDAllBlue();
          if (LoopCount > 2)
          {
            goto __ExitWaitChoice;
          }
          break;
        case 1:
          LEDAllRed();
          break;
        case 2:
          LEDAllGreen();
          break;
        case 3:
          LEDAllBlue();
          LEDAllRed();
          break;
      }
#else
      switch (Option)
      {
        case 0:
          LoopCount++;
          gpio_direction_output(LED_LAN_G, 0);
          gpio_direction_output(LED_WLAN_G, 0);
          if (LoopCount > 2)
          {
            goto __ExitWaitChoice;
          }
          break;
        case 1:
          gpio_direction_output(LED_LAN_G, 0);
          gpio_direction_output(LED_WLAN_R, 0);
          break;
        case 2:
          gpio_direction_output(LED_LAN_R, 0);
          gpio_direction_output(LED_WLAN_G, 0);
          break;
        case 3:
          gpio_direction_output(LED_LAN_R, 0);
          gpio_direction_output(LED_WLAN_R, 0);
          break;
      }
#endif
      for (i = 0; i < 1000; i++)
      {
        if (gpio_get_value(GPIO_RESET)) break;
        udelay(1000);
      }
    }
    printf("choice: %d\r\n", Option);

    __ExitWaitChoice:
    if (Option == 1)
    {
      LEDAllOff();
      LEDAllRed();
      if (LoopCount > 1)
      {
        setenv("serverip", "192.168.170." SERVERIP);
        setenv("ipaddr", "192.168.170.111");
      } else
      {
        setenv("serverip", "192.168.190." SERVERIP);
        setenv("ipaddr", "192.168.190.111");
      }
      setenv("bootcmd", "run BootTFTP");
      setenv("bootdelay", "0");
    } else if (Option == 0)
    {
      if (LoopCount)
      {
        sprintf(&TmpBootArgs[strlen(TmpBootArgs)], " lc=%d ", LoopCount);
        setenv("bootargs", TmpBootArgs);
      }
      setenv("bootdelay", "2");
    } else if (Option == 2)
    {
      while (1);
#if 0
      int RC;
      RC = mmdc_do_write_level_calibration();
      RC = mmdc_do_dqs_calibration();
      printf("DDR3: Completed RC = %08X\r\n", RC);
      setenv("bootcmd", "");
#endif
    } else if (Option == 3)
    {
      ExecuteUCP = 1;
      setenv("bootcmd", "");
    }
  } else
  {
    setenv("bootdelay", "0");
  }

  printf("go...\r\n");

  return 0;
}

#ifdef CONFIG_CMD_BMODE
static const struct boot_mode board_boot_modes[] = {
	/* 4 bit bus width */
	{"sd3",	 MAKE_CFGVAL(0x40, 0x30, 0x00, 0x00)},
	{NULL,	 0},
};
#endif


int board_late_init(void)
{
#ifdef CONFIG_CMD_BMODE
	add_board_boot_modes(board_boot_modes);
#endif

	return 0;
}


#pragma pack(1)

typedef struct SystemDataTLV_s
{
  __uint16_t Type;
  __uint16_t Length;
  union
  {
    __uint8_t UInt8;
    __uint16_t UInt16;
    __uint32_t UInt32;
    __uint64_t UInt64;
    char String[8];
    __uint8_t Binary[8];
  } Value;
  char AlwaysNull;
} SystemDataTLV_t;

#define SYSTEM_DATA_SERNUM    0x0001
#define SYSTEM_DATA_MAGIC     0xFA5F

char SystemDataMagic[] = { 0x5F, 0xFA, 0x08, 0x00, 0x41, 0x54, 0x41, 0x44, 0x5F, 0x53, 0x59, 0x53, };

#pragma pack()

int InitSystemData(void)
{
  struct spi_flash *Flash = spi_flash_probe(CONFIG_SF_DEFAULT_BUS, CONFIG_SF_DEFAULT_CS, CONFIG_SF_DEFAULT_SPEED, CONFIG_SF_DEFAULT_MODE);
  if (Flash)
  {
    int Offset = 0x000C0000;
    char TmpMagic[sizeof(SystemDataMagic)];
    spi_flash_read(Flash, Offset, sizeof(TmpMagic), &TmpMagic);
    if (memcmp(SystemDataMagic, TmpMagic, sizeof(TmpMagic)) == 0)
    {
      printf("SystemData: SPI[0x%08X]: ", Offset);
    }

    int LimitElements = 100;
    while (LimitElements--)
    {
      __uint16_t Type;
      __uint16_t Length;
      spi_flash_read(Flash, Offset, sizeof(Type), &Type);
      Offset += 2;
      spi_flash_read(Flash, Offset, sizeof(Length), &Length);
      if (!Type || !Length || Type == 0xFFFF || Length == 0xFFFF) break;
      // printf("SysData: %04X %04X\r\n", Type, Length);
      Offset += 2;
      switch (Type)
      {
        case SYSTEM_DATA_SERNUM:
        {
          __uint64_t SerNum;
          spi_flash_read(Flash, Offset, sizeof(SerNum), &SerNum);
          printf("Sernum: %d ", (int) SerNum);
          break;
        }
        case SYSTEM_DATA_MAGIC:
        {
          __uint64_t Magic;
          spi_flash_read(Flash, Offset, sizeof(Magic), &Magic);
          if (Magic == 0x5359535F44415441)
          {
            printf("Found Magic! ");
            FoundMagic = 1;
          }
          break;
        }
      }
      Offset += Length + 1;
    }
    printf("\r\n");
  } else
  {
    struct mmc *MMCDev = find_mmc_device(CONFIG_SYS_MMC_ENV_DEV);
    if (init_mmc_for_env(MMCDev)) return 0;
    char *Buffer = (char *) 0x12000000;
    int Offset = 0x000C0000;
    uint BlkStart, BlkCount, n;
    BlkStart = ALIGN(Offset, MMCDev->write_bl_len) / MMCDev->write_bl_len;
    BlkCount = 1;
    printf("Read from MMC offset %08X BlockCount %08X!\n", BlkStart, BlkCount);
    n = MMCDev->block_dev.block_read(CONFIG_SYS_MMC_ENV_DEV, BlkStart, BlkCount, (u_char *) Buffer);
    if (n != BlkCount)
    {
      printf("Read from MMC Failed!\n");
    } else
    {
      FoundMMC = 1;
      __uint64_t Magic;
      memcpy(&Magic, Buffer + 4, sizeof(__uint64_t));
      if (Magic == 0x5359535F44415441)
      {
        printf("Found Magic in MMC!\n");
        FoundMagic = 1;
      }
    }

    if (FoundMMC)
    {
      BlkStart = 2;
      BlkCount = 1;
      printf("Read from MMC offset %08X BlockCount %08X!\n", BlkStart, BlkCount);
      n = MMCDev->block_dev.block_read(CONFIG_SYS_MMC_ENV_DEV, BlkStart, BlkCount, (u_char *) Buffer);
      if (n != BlkCount)
      {
        printf("Read from MMC Failed!\n");
      } else
      {
        if (Buffer[0] != 0x00 || Buffer[1] != 0x00 || Buffer[2] != 0x00 || Buffer[3] != 0x00)
        {
          printf("Found uboot! %02X%02X%02X%02X\n", Buffer[0], Buffer[1], Buffer[2], Buffer[3]);
          FoundMMCUBoot = 1;
        }
      }
    }
  }
  return 0;
}

int last_stage_init(void)
{
  if (ExecuteUCP)
  {
    setenv("ipaddr", "192.168.0.1");
    setenv("serverip", "192.168.0.1");
    NetStartAgain();
    NetLoop(UCP);
  }

  return 0;
}

int checkboard(void)
{
	puts("Board: Modas MC1\n");

	return 0;
}
