/*
 * Copyright (C) 2012 Freescale Semiconductor, Inc.
 *
 * Author: Fabio Estevam <fabio.estevam@freescale.com>
 *
 * See file CREDITS for list of people who contributed to this
 * project.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 */

#include <common.h>
#include <asm/io.h>
#include <asm/arch/clock.h>
#include <asm/arch/imx-regs.h>
#include <asm/arch/iomux.h>
#include <asm/arch/mx6-pins.h>
#include <asm/errno.h>
#include <asm/gpio.h>
#include <asm/imx-common/iomux-v3.h>
#include <asm/imx-common/boot_mode.h>
#include <mmc.h>
#include <fsl_esdhc.h>
#include <miiphy.h>
#include <netdev.h>
#include <asm/imx-common/mxc_i2c.h>
#include <i2c.h>

// #include "ddrcal.h"
#include "mx6smodas.h"

int mmdc_do_write_level_calibration_ocram(void);
int mmdc_do_dqs_calibration_ocram(void);

typedef int (*DDRCalibFunction_t) (void);

static int PRandState = 1234;
void PRandInit(int NewState)
{
  PRandState = NewState;
}

int PRand(void)
{
  PRandState = ((PRandState * 1103515245) + 12345) & 0x7fffffff;
  return PRandState;
}

void hexdump(void *Data, int DataLen)
{
  int i = 0;
  char *DP = (char *) Data;
  while (DataLen)
  {
    printf("%02X", DP[i]);
    i++;
    if (i % 16 == 0)
    {
      printf("\r\n");
    }
    DataLen--;
  }
}


#define P0_IPS (void __iomem *)0x021b0000
#define P1_IPS (void __iomem *)0x021b4000


#define MDCTL   0x000
#define MDPDC   0x004
#define MDSCR   0x01c
#define MDMISC    0x018
#define MDREF   0x020
#define MAPSR   0x404
#define MPZQHWCTRL  0x800
#define MPWLGCR   0x808
#define MPWLDECTRL0 0x80c
#define MPWLDECTRL1 0x810
#define MPPDCMPR1 0x88c
#define MPSWDAR   0x894
#define MPRDDLCTL 0x848
#define MPMUR   0x8b8
#define MPDGCTRL0 0x83c
#define MPDGCTRL1 0x840
#define MPRDDLCTL 0x848
#define MPWRDLCTL 0x850
#define MPRDDLHWCTL 0x860
#define MPWRDLHWCTL 0x864
#define MPDGHWST0 0x87c
#define MPDGHWST1 0x880
#define MPDGHWST2 0x884
#define MPDGHWST3 0x888
int mmdc_do_write_level_calibration(void)
{
  char Test = 0;
  DDRCalibFunction_t Target = (void *) (IRAM_BASE_ADDR + IRAM_SIZE/2);
  printf("mmdc_do_write_level_calibration_ocram: call at %p -> (Stack: %p)\r\n", Target, &Test);
  memcpy(Target, mmdc_do_write_level_calibration_ocram, 4096);
  hexdump(Target, 64);
  int RC = Target();


  printf("MMDC_MPWLDECTRL0 after write level cal: 0x%08X\n",
      readl(P0_IPS + MPWLDECTRL0));
  printf("MMDC_MPWLDECTRL1 after write level cal: 0x%08X\n",
      readl(P0_IPS + MPWLDECTRL1));

  return RC;
}

int mmdc_do_dqs_calibration(void)
{
  char Test = 0;
  DDRCalibFunction_t Target = (void *) (IRAM_BASE_ADDR + IRAM_SIZE/2);
  printf("mmdc_do_dqs_calibration_ocram: call at %p -> (Stack: %p)\r\n", Target, &Test);
  memcpy(Target, mmdc_do_dqs_calibration_ocram, 4096);
  hexdump(Target, 64);
  int RC = Target();
  return RC;
}


DECLARE_GLOBAL_DATA_PTR;

iomux_v3_cfg_t const led_pads[] =
{
 MX6_PAD_NANDF_CS1__GPIO6_IO14 | MUX_PAD_CTRL(LED_PAD_CTRL),
 MX6_PAD_NANDF_CS2__GPIO6_IO15 | MUX_PAD_CTRL(LED_PAD_CTRL),
 MX6_PAD_NANDF_CS3__GPIO6_IO16 | MUX_PAD_CTRL(LED_PAD_CTRL),
 MX6_PAD_NANDF_ALE__GPIO6_IO08  | MUX_PAD_CTRL(LED_PAD_CTRL),
 MX6_PAD_NANDF_CLE__GPIO6_IO07  | MUX_PAD_CTRL(LED_PAD_CTRL),
 MX6_PAD_NANDF_WP_B__GPIO6_IO09 | MUX_PAD_CTRL(LED_PAD_CTRL),
 MX6_PAD_NANDF_RB0__GPIO6_IO10 | MUX_PAD_CTRL(LED_PAD_CTRL),
};

void led_setup(void)
{
  imx_iomux_v3_setup_multiple_pads(led_pads, ARRAY_SIZE(led_pads));

  gpio_direction_output(LED1_R, 0);
  gpio_direction_output(LED1_G, 0);
  gpio_direction_output(LED2_R, 0);
  gpio_direction_output(LED2_G, 0);
  gpio_direction_output(LED3_R, 0);
  gpio_direction_output(LED3_G, 0);
  gpio_direction_output(LED4_R, 0);
}

int dram_init(void)
{
	gd->ram_size = get_ram_size((void *)PHYS_SDRAM, PHYS_SDRAM_SIZE);
	return 0;
}

/* I2C2 PMIC, iPod, Tuner, Codec, Touch, HDMI EDID, MIPI CSI2 card */
struct i2c_pads_info i2c_pad_info1 = {
  .scl = {
    .i2c_mode = MX6_PAD_KEY_COL3__I2C2_SCL | MUX_PAD_CTRL(I2C_PAD_CTRL),
    .gpio_mode = MX6_PAD_KEY_COL3__GPIO4_IO12 | MUX_PAD_CTRL(I2C_PAD_CTRL),
    .gp = IMX_GPIO_NR(4, 12)
  },
  .sda = {
    .i2c_mode = MX6_PAD_KEY_ROW3__I2C2_SDA | MUX_PAD_CTRL(I2C_PAD_CTRL),
    .gpio_mode = MX6_PAD_KEY_ROW3__GPIO4_IO13 | MUX_PAD_CTRL(I2C_PAD_CTRL),
    .gp = IMX_GPIO_NR(4, 13)
  }
};


iomux_v3_cfg_t const uart1_pads[] = {
	MX6_PAD_CSI0_DAT10__UART1_TX_DATA | MUX_PAD_CTRL(UART_PAD_CTRL),
	MX6_PAD_CSI0_DAT11__UART1_RX_DATA | MUX_PAD_CTRL(UART_PAD_CTRL),
};


/*
 * 4-Bit Modus
 * Kein Card-Detect
 */
iomux_v3_cfg_t const usdhc3_pads[] =
{
	MX6_PAD_SD3_CLK__SD3_CLK      | MUX_PAD_CTRL(USDHC_PAD_CTRL),
	MX6_PAD_SD3_CMD__SD3_CMD      | MUX_PAD_CTRL(USDHC_PAD_CTRL),
	MX6_PAD_SD3_DAT0__SD3_DATA0   | MUX_PAD_CTRL(USDHC_PAD_CTRL),
	MX6_PAD_SD3_DAT1__SD3_DATA1   | MUX_PAD_CTRL(USDHC_PAD_CTRL),
	MX6_PAD_SD3_DAT2__SD3_DATA2   | MUX_PAD_CTRL(USDHC_PAD_CTRL),
	MX6_PAD_SD3_DAT3__SD3_DATA3   | MUX_PAD_CTRL(USDHC_PAD_CTRL),
};

static void setup_iomux_uart(void)
{
	imx_iomux_v3_setup_multiple_pads(uart1_pads, ARRAY_SIZE(uart1_pads));
}

struct fsl_esdhc_cfg usdhc_cfg[1] =
{
	{USDHC3_BASE_ADDR},
};

#define GPIO_RESET     IMX_GPIO_NR(1, 2)


#define USBOTG_IN_PAD_CTRL  (PAD_CTL_HYS | PAD_CTL_PUS_100K_DOWN | PAD_CTL_DSE_240ohm | PAD_CTL_SRE_SLOW)
#define USBOTG_OUT_PAD_CTRL (                                      PAD_CTL_DSE_240ohm | PAD_CTL_SRE_SLOW)

static void setup_iomux_usbotg(void)
{
  static const iomux_v3_cfg_t usbotg_pads[] =
  {
   MX6_PAD_EIM_D22__USB_OTG_PWR  | MUX_PAD_CTRL(USBOTG_OUT_PAD_CTRL),
   MX6_PAD_EIM_D21__USB_OTG_OC  | MUX_PAD_CTRL(USBOTG_IN_PAD_CTRL),
   MX6_PAD_ENET_RX_ER__USB_OTG_ID  | MUX_PAD_CTRL(USBOTG_IN_PAD_CTRL),
  };

  /* Set up pins for USBOTG. */
  imx_iomux_v3_setup_multiple_pads(usbotg_pads, ARRAY_SIZE(usbotg_pads));
}

int board_early_init_f(void)
{
  led_setup();
  setup_iomux_usbotg();
	setup_iomux_uart();
  printf("Debug...\n");
	return 0;
}

iomux_v3_cfg_t const ecspi3_pads[] =
{
 MX6_PAD_DISP0_DAT3__GPIO4_IO24  | MUX_PAD_CTRL(SPI_PAD_CTRL), /* SS0 */
 MX6_PAD_DISP0_DAT2__ECSPI3_MISO | MUX_PAD_CTRL(SPI_PAD_CTRL),
 MX6_PAD_DISP0_DAT1__ECSPI3_MOSI | MUX_PAD_CTRL(SPI_PAD_CTRL),
 MX6_PAD_DISP0_DAT0__ECSPI3_SCLK | MUX_PAD_CTRL(SPI_PAD_CTRL),
};

void setup_spi(void)
{
  puts("SPI: Init pad configuration (ECSPI3, CS0)\n");
  imx_iomux_v3_setup_multiple_pads(ecspi3_pads, ARRAY_SIZE(ecspi3_pads));
}

int board_init(void)
{
	/* address of boot parameters */
	gd->bd->bi_boot_params = PHYS_SDRAM + 0x100;

  return 0;
}

int misc_init_r(void)
{

  return 0;
}

#ifdef CONFIG_CMD_BMODE
static const struct boot_mode board_boot_modes[] = {
	/* 4 bit bus width */
	{"sd3",	 MAKE_CFGVAL(0x40, 0x30, 0x00, 0x00)},
	{NULL,	 0},
};
#endif


int board_late_init(void)
{
#ifdef CONFIG_CMD_BMODE
	add_board_boot_modes(board_boot_modes);
#endif

	return 0;
}

int checkboard(void)
{
	puts("Board: Modas MC1\n");

	return 0;
}
