/*
 * ucp.c
 *
 *  Created on: 03.02.2014
 *      Author: thomas
 */

#include <common.h>
#include <net.h>

#pragma pack(1)

typedef unsigned char __uint8_t;
typedef unsigned short __uint16_t;
typedef unsigned int __uint32_t;

// Flags: 0x1 = Request
typedef struct __attribute__ ((packed)) UCPRequest_s
{
  __uint8_t DestAddr[8];
  __uint8_t SrcAddr[8];
  __uint16_t Seq;
  __uint16_t Type;
  __uint8_t Flags;
  __uint32_t Interface;
  __uint16_t Message;
  __uint8_t Data[2048];
} UCPRequest_t;

// Flags: 0x0 = Reply
typedef struct __attribute__ ((packed)) UCPReply_s
{
  __uint8_t DestAddr[8];
  __uint8_t SrcAddr[8];
  __uint16_t Seq;
  __uint16_t Type;
  __uint8_t Flags;
  __uint32_t Interface;
  __uint16_t Message;
  __uint8_t Data[2048];
} UCPReply_t;
#pragma pack()

#ifndef CONFIG_UCP_UPG_ADDR
#error Address for firmware upgrade needed.
#endif

static __uint8_t *AddTLV(__uint8_t *Reply, unsigned char ID, unsigned char Len, void *Data)
{
  if (!Reply) return NULL;
  *Reply++ = ID;
  *Reply++ = Len;
  memcpy(Reply, Data, Len);
  Reply += Len;
  return Reply;
}

static void UCPSend(UCPReply_t *Reply, __uint16_t RemotePort, size_t Len)
{
//  printf("%s %d Bytes\n", __func__, Len);
  uchar *TxPkt;
  TxPkt = NetTxPacket + NetEthHdrSize() + IP_UDP_HDR_SIZE;
  memcpy(TxPkt, Reply, Len);
  NetSendUDPPacket((uchar *) "\xFF\xFF\xFF\xFF\xFF\xFF", 0xFFFFFFFF, RemotePort, 17784, Len);
}


void UCPExecuteFirmwareUpgrade(void *Addr, size_t Len) __attribute__((weak, alias("UCPExecuteFirmwareUpgradeDefault")));

static void UCPExecuteFirmwareUpgradeDefault(void *Addr, size_t Len)
{
  printf("UCPExecuteFirmwareUpgrade Function missing! %p %d\n", Addr, Len);
}

/*
 *  Handle DHCP received packets.
 */
static void UCPHandler(uchar *pkt, unsigned dest, IPaddr_t sip, unsigned src,
      unsigned len)
{
  static size_t LocalUpgradeOffset = 0;
  static size_t NextSeqNum = 0;
  static UCPRequest_t __attribute__ ((aligned (16))) UcpReq;
  static UCPReply_t __attribute__ ((aligned (16))) UcpRep;
  // printf("UCPHandler: got UCP packet: (src=%pI4:%d, dst=%d, len=%d)\n", &sip, src, dest, len);
  if (pkt)
  {
    memset(&UcpRep, 0, sizeof(UcpRep));
    memset(&UcpReq, 0, sizeof(UcpReq));
    memcpy(&UcpReq, pkt, len);

    if (UcpReq.Flags & 0x01)
    {
      /*
       * Reply Header aufbauen - Kopie, nur mit Getauschten MACs
       */
      memcpy(UcpRep.DestAddr,     UcpReq.SrcAddr, 8);
      memcpy(&UcpRep.SrcAddr[0],  "\x00\x01", 2);
      memcpy(&UcpRep.SrcAddr[2],  NetOurEther, 6);
      memcpy(&UcpRep.Seq,         &UcpReq.Seq, 2);
      memcpy(&UcpRep.Type,        &UcpReq.Type, 2);
      memcpy(&UcpRep.Interface,   &UcpReq.Interface, 4);
      memcpy(&UcpRep.Message,     &UcpReq.Message, 2);

#if 0
      int i;
      printf("UCP:");
      for (i = 0; i < len; i++)
        printf("%02X ", pkt[i]);
      printf("\n");
#endif
      __uint16_t Message;
      memcpy(&Message, &UcpReq.Message, 2);
      // printf("UCP Method: %d\n", htons(Message));
      __uint8_t *UcpData = UcpRep.Data;

      switch (htons(Message))
      {
        case 1:
        {
          char DevType = 99;
          UcpData = AddTLV(UcpData, 0x03, 1, &DevType);
          UcpData = AddTLV(UcpData, 0x02, 16, "U-BOOT Recovery");
          break;
        }
#ifdef CONFIG_UCP_UPG_ADDR
        case 12: // Upgrade...
        {
          // char *UpgradeDataP = &UcpReq.Data[34];
          memcpy(UcpData, &UcpReq.Data[32], 2);
          UcpData += 2;

          __uint16_t Seq;
          memcpy(&Seq, &UcpReq.Data[32], 2);
          if (!Seq || htons(Seq) >= 9999)
          {
            printf("Seq: %d %08X\n", htons(Seq), LocalUpgradeOffset);
          }
          if (!Seq)
          {
            LocalUpgradeOffset = CONFIG_UCP_UPG_ADDR;
            NextSeqNum = 0;
          }

          if (htons(Seq) == NextSeqNum)
          {
            int CurrentDataLen = len - 27 - 34;
            memcpy((void *) LocalUpgradeOffset, &UcpReq.Data[34], CurrentDataLen);
            LocalUpgradeOffset += CurrentDataLen;
            NextSeqNum++;
          }
          if (htons(Seq) == 9999)
          {
            UCPExecuteFirmwareUpgrade((void *) CONFIG_UCP_UPG_ADDR, LocalUpgradeOffset - CONFIG_UCP_UPG_ADDR);
          }
          break;
        }
#endif
      }
      if (UcpData != UcpRep.Data)
      {
        UCPSend(&UcpRep, src, offsetof(UCPReply_t, Data) + ((size_t) UcpData - (size_t) UcpRep.Data));
      }
    }
  }
}

void ucp_start(void)
{
  net_set_udp_handler(UCPHandler);
}


